# RCAIDE/Methods/Energy/Propulsors/Constant_Speed_ICE_Propulsor/compute_cs_ice_performance.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports  
from RCAIDE.Framework.Core import Units  
from RCAIDE.Library.Methods.Propulsors.Converters.Engine import compute_throttle_from_power
from RCAIDE.Library.Methods.Propulsors.Converters.Rotor.compute_rotor_performance import  compute_rotor_performance
 
# pacakge imports  
from copy import deepcopy
import numpy as np  

# ----------------------------------------------------------------------------------------------------------------------
# internal_combustion_engine_constant_speed_propulsor
# ----------------------------------------------------------------------------------------------------------------------  
def compute_cs_ice_performance(propulsor,state,center_of_gravity= [[0.0, 0.0,0.0]]):  
    ''' Computes the perfomrance of one propulsor
    
    Assumptions: 
    N/A

    Source:
    N/A

    Inputs:  
    conditions           - operating conditions data structure           [-]  
    fuel_line            - fuelline                                      [-] 
    propulsor        - propulsor data structure            [-] 
    total_thrust         - thrust of propulsor group              [N]
    total_power          - power of propulsor group               [W] 

    Outputs:  
    total_thrust         - thrust of propulsor group              [N]
    total_power          - power of propulsor group               [W] 
    stored_results_flag  - boolean for stored results                    [-]     
    stored_propulsor_tag - name of propulsor with stored results  [-]
    
    Properties Used: 
    N.A.        
    '''  
    conditions              = state.conditions  
    ice_cs_conditions       = conditions.energy[propulsor.tag] 
    engine                  = propulsor.engine 
    propeller               = propulsor.propeller
    engine_conditions       = ice_cs_conditions[engine.tag]
    engine_conditions.rpm   = conditions.energy[propulsor.tag].rpm 

    # Run the propeller to get the power
    propeller_conditions                = ice_cs_conditions[propeller.tag]
    propeller_conditions.omega          = engine_conditions.rpm * Units.rpm
    propeller_conditions.pitch_command  = ice_cs_conditions.throttle - 0.5
    propeller_conditions.throttle       = ice_cs_conditions.throttle
    compute_rotor_performance(propulsor,state,center_of_gravity)

    # Run the engine to calculate the throttle setting and the fuel burn
    engine_conditions.power        = conditions.energy[propulsor.tag][propeller.tag].power 
    compute_throttle_from_power(engine,engine_conditions,conditions) 
    
    # Create the outputs
    ice_cs_conditions.fuel_flow_rate         = engine_conditions.fuel_flow_rate  
    stored_results_flag                      = True
    stored_propulsor_tag                     = propulsor.tag  

    # compute total forces and moments from propulsor (future work would be to add moments from motors)
    conditions.energy[propulsor.tag].thrust      = conditions.energy[propulsor.tag][propeller.tag].thrust 
    conditions.energy[propulsor.tag].moment      = conditions.energy[propulsor.tag][propeller.tag].moment
    conditions.energy[propulsor.tag].power       = conditions.energy[propulsor.tag][propeller.tag].power 
    T  = conditions.energy[propulsor.tag].thrust 
    M  = conditions.energy[propulsor.tag].moment 
    P  = conditions.energy[propulsor.tag].power 
    
    return T,M,P,stored_results_flag,stored_propulsor_tag 
    
def reuse_stored_ice_cs_prop_data(propulsor,state,network,stored_propulsor_tag,center_of_gravity= [[0.0, 0.0,0.0]]):
    '''Reuses results from one propulsor for identical propulsors
    
    Assumptions: 
    N/A

    Source:
    N/A

    Inputs:  
    conditions           - operating conditions data structure    [-]  
    fuel_line            - fuel_line                              [-] 
    propulsor        - propulsor data structure     [-] 
    total_thrust         - thrust of propulsor group       [N]
    total_power          - power of propulsor group        [W] 
     
    Outputs:      
    total_thrust         - thrust of propulsor group       [N]
    total_power          - power of propulsor group        [W] 
    
    Properties Used: 
    N.A.        
    ''' 
    conditions                 = state.conditions
    engine                     = propulsor.engine
    propeller                  = propulsor.propeller 
    engine_0                   = network.propulsors[stored_propulsor_tag].engine
    propeller_0                = network.propulsors[stored_propulsor_tag].propeller

    conditions.energy[propulsor.tag][engine.tag]        = deepcopy(conditions.energy[stored_propulsor_tag][engine_0.tag])
    conditions.energy[propulsor.tag][propeller.tag]     = deepcopy(conditions.energy[stored_propulsor_tag][propeller_0.tag])    
  
    thrust                  = conditions.energy[propulsor.tag][propeller.tag].thrust 
    power                   = conditions.energy[propulsor.tag][propeller.tag].power 
    
    moment_vector           = 0*state.ones_row(3) 
    moment_vector[:,0]      = propeller.origin[0][0]  -  center_of_gravity[0][0] 
    moment_vector[:,1]      = propeller.origin[0][1]  -  center_of_gravity[0][1] 
    moment_vector[:,2]      = propeller.origin[0][2]  -  center_of_gravity[0][2]
    moment                  =  np.cross(moment_vector, thrust)
    
    conditions.energy[propulsor.tag][propeller.tag].moment = moment  
    conditions.energy[propulsor.tag].thrust            = thrust   
    conditions.energy[propulsor.tag].moment            = moment  
    conditions.energy[propulsor.tag].power             = power
 
    return thrust,moment,power
 


# RCAIDE/Library/Missions/Common/Update/curvilinear_inertial_horizontal_position.py
# 
# 
# Created:  September 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# Package imports 
import numpy as np
from RCAIDE.Framework.Core import Units   

# ----------------------------------------------------------------------------------------------------------------------
#  Integrate Position
# ---------------------------------------------------------------------------------------------------------------------- 
def curvilinear_inertial_horizontal_position(segment):
    """ Determines how far the airplane has traveled and calculates position. 
    
        Assumptions:
        Assumes a flat earth, this is planar motion.
        
        Inputs:
            segment.state.conditions:
                frames.inertial.position_vector [meters]
                frames.inertial.velocity_vector [meters/second]
            segment.state.numerics.time.integrate       [float]
            
        Outputs:
            segment.state.conditions:           
                frames.inertial.position_vector [meters]

        Properties Used:
        N/A
                                
    """        

    conditions  = segment.state.conditions 
    psi         = conditions.frames.planet.true_heading
    x0          = conditions.frames.inertial.position_vector[0, 0] 
    y0          = conditions.frames.inertial.position_vector[0, 1]
    R0          = conditions.frames.inertial.aircraft_range[0,None,0:1+1]
    vx          = conditions.frames.inertial.velocity_vector[:,0:1+1]
    I           = segment.state.numerics.time.integrate 
    R           = segment.turn_radius
    sign        = np.sign(segment.turn_angle)
    
    # integrate
    speed       = np.sqrt(vx[:, 0]**2+vx[:, 1]**2)
    arc_length  = np.dot(I,speed)
    
    theta       = psi - sign * 90 * Units.degrees         # Angle from circle center to the flight trajectory
    beta        =  psi[0, 0] + sign * 90 * Units.degrees  # Angle to the center of the circle from the initial position
    
    delta_x     = R * np.cos(beta) + R * np.cos(theta) # vector addition with a vector from the starting point to the center and then from the center to the position
    delta_y     = R * np.sin(beta) + R * np.sin(theta)
    x_position  = x0 + delta_x
    y_position  = y0 + delta_y
    
    # pack
    conditions.frames.inertial.position_vector[:,0] = x_position[:,0]
    conditions.frames.inertial.position_vector[:,1] = y_position[:,0] 
    conditions.frames.inertial.aircraft_range[:,0]  = R0 + arc_length
    return


# RCAIDE/Library/Methods/Weights/Buildups/eVTOL/compute_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
import RCAIDE
from RCAIDE.Framework.Core                                          import Units, Data 
from RCAIDE.Library.Methods.Weights.Physics_Based_Buildups.Common   import compute_fuselage_weight
from RCAIDE.Library.Methods.Weights.Physics_Based_Buildups.Common   import compute_boom_weight
from RCAIDE.Library.Methods.Weights.Physics_Based_Buildups.Common   import compute_rotor_weight
from RCAIDE.Library.Methods.Weights.Physics_Based_Buildups.Common   import compute_wiring_weight
from RCAIDE.Library.Methods.Weights.Physics_Based_Buildups.Common   import compute_wing_weight

# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
# Compute Operating Empty Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_operating_empty_weight(vehicle,settings = None): 
    """ Calculates the empty vehicle mass for an EVTOL-type aircraft including seats,
        avionics, servomotors, ballistic recovery system, rotor and hub assembly,
        transmission, and landing gear. 
        
        Sources:
        Project Vahana Conceptual Trade Study
        https://github.com/VahanaOpenSource


        Inputs: 
            vehicle:                     RCAIDE Config Data Stucture 
            max_tip_mach:               Allowable Tip Mach Number                      [Unitless]
            disk_area_factor:           Inverse of Disk Area Efficiency                [Unitless]
            max_thrust_to_weight_ratio: Allowable Thrust to Weight Ratio               [Unitless]
            safety_factor               Safety Factor in vehicle design                [Unitless]
            max_g_load                  Maximum g-forces load for certification        [UNitless]
            motor_efficiency:           Motor Efficiency                               [Unitless]

        Outputs: 
            outputs:                    Data Dictionary of Component Masses [kg]

        Output data dictionary has the following book-keeping hierarchical structure:

            Output
                Total.
                    Empty.
                        Structural.
                            Fuselage
                            Wings
                            Landing Gear
                            Rotors
                            Hubs
                        Seats
                        Battery
                        Motors
                        Servo
                    Systems.
                        Avionics
                        ECS               - Environmental Control System
                        BRS               - Ballistic Recovery System
                        Wiring            - Aircraft Electronic Wiring
                    Payload

    """
    
    if settings == None: 
        miscelleneous_weight_factor   = 1.1
        safety_factor                 = 1.5   
        disk_area_factor              = 1.15     
        max_thrust_to_weight_ratio    = 1.1
        max_g_load                    = 3.8
    else:
        miscelleneous_weight_factor   = settings.miscelleneous_weight_factor
        safety_factor                 = settings.safety_factor              
        disk_area_factor              = settings.disk_area_factor           
        max_thrust_to_weight_ratio    = settings.max_thrust_to_weight_ratio 
        max_g_load                    = settings.max_g_load  
               
    # Set up data structures for RCAIDE weight methods
    weight                                  = Data()  
    weight.battery                          = 0.0
    weight.payload                          = 0.0
    weight.servos                           = 0.0
    weight.hubs                             = 0.0
    weight.booms                            = 0.0
    weight.BRS                              = 0.0  
    weight.motors                           = 0.0
    weight.rotors                           = 0.0
    weight.rotors                           = 0.0
    weight.fuselage                         = 0.0
    weight.wiring                           = 0.0
    weight.wings                            = Data()
    weight.wings_total                      = 0.0
    weight.thermal_management_system       = Data()


    control_systems                                  = RCAIDE.Library.Components.Component()
    control_systems.tag                              = 'control_systems'  
    electrical_systems                               = RCAIDE.Library.Components.Component()
    electrical_systems.tag                           = 'electrical_systems'
    furnishings                                      = RCAIDE.Library.Components.Component()
    furnishings.tag                                  = 'furnishings'
    air_conditioner                                  = RCAIDE.Library.Components.Component() 
    air_conditioner.tag                              = 'air_conditioner'
    apu                                              = RCAIDE.Library.Components.Component()
    apu.tag                                          = 'apu'
    hydraulics                                       = RCAIDE.Library.Components.Component()
    hydraulics.tag                                   = 'hydraulics'
    avionics                                         = RCAIDE.Library.Components.Systems.Avionics()
    optionals                                        = RCAIDE.Library.Components.Component()
    optionals.tag                                    = 'optionals'
    

    vehicle.payload.passengers      = RCAIDE.Library.Components.Component()
    vehicle.payload.passengers.tag  = 'passengers'  
    vehicle.payload.baggage         = RCAIDE.Library.Components.Component()
    vehicle.payload.baggage.tag     = 'baggage'
    vehicle.payload.cargo           = RCAIDE.Library.Components.Component()   
    vehicle.payload.cargo.tag       = 'cargo'

    # assign components to vehicle
    vehicle.systems.control_systems    = control_systems
    vehicle.systems.electrical_systems = electrical_systems
    vehicle.systems.avionics           = avionics
    vehicle.systems.furnishings        = furnishings
    vehicle.systems.air_conditioner    = air_conditioner 
    vehicle.systems.apu                = apu
    vehicle.systems.hydraulics         = hydraulics
    vehicle.systems.optionals          = optionals 

    #-------------------------------------------------------------------------------
    # Default Values
    #------------------------------------------------------------------------------- 
    MTOW                          = vehicle.mass_properties.max_takeoff
    atmosphere                    = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976() 
    atmo_data                     = atmosphere.compute_values(0, 0) 
    rho_ref                       = atmo_data.density[0, 0] 
    maxLift                       = MTOW * max_thrust_to_weight_ratio * 9.81           
    AvgBladeCD                    = 0.012         

    #-------------------------------------------------------------------------------
    # Fixed Weights
    #-------------------------------------------------------------------------------
    weight.seats        = vehicle.passengers * 15.   * Units.kg
    weight.passengers   = vehicle.passengers * 70.   * Units.kg
    weight.avionics     = 15.                        * Units.kg
    weight.landing_gear = MTOW * 0.02                * Units.kg
    weight.ECS          = vehicle.passengers * 7.    * Units.kg

    # Determine length scale 
    length_scale = 1.  
    if len(vehicle.fuselages) == 0.:
        for wing in vehicle.wings:
            if isinstance(wing ,RCAIDE.Library.Components.Wings.Main_Wing): 
                if wing.chords.root>length_scale:
                    length_scale = wing.chords.root
                    nose_length  = 0.25*wing.chords.root
    else:
        for fuse in vehicle.fuselages:
            nose   = fuse.lengths.nose
            length = fuse.lengths.total
            if length > length_scale:
                length_scale = length
                nose_length  = nose 
                            
    #-------------------------------------------------------------------------------
    # Environmental Control System
    #-------------------------------------------------------------------------------
    vehicle.systems.air_conditioner.origin[0][0]          = 0.51 * length_scale
    vehicle.systems.air_conditioner.mass_properties.mass  = weight.ECS

    #-------------------------------------------------------------------------------
    # Network Weight
    #-------------------------------------------------------------------------------
    maxLiftPower           = 0
    total_number_of_rotors = 0
    maxVTip                = 0
    eta                    = 0
    for network in vehicle.networks: 
        for bus in network.busses:  
            #------------------------------------------------------------------------------- 
            # Payload Weight
            #-------------------------------------------------------------------------------
            if bus.payload.origin[0][0] == 0:
                bus.payload.origin[0][0]  = 0.5 * length_scale
            weight.payload  += bus.payload.mass_properties.mass * Units.kg
     
            #-------------------------------------------------------------------------------
            # Avionics Weight
            #-------------------------------------------------------------------------------
            if bus.avionics.origin[0][0] == 0:
                bus.avionics.origin[0][0]                          = 0.4 * nose_length
            bus.avionics.mass_properties.center_of_gravity[0][0]   = 0.0
            weight.avionics += bus.avionics.mass_properties.mass   
                         
            for modules in bus.battery_modules: 
                weight.battery += modules.mass_properties.mass * Units.kg  
        
            # Servo, Hub and BRS Weights
            lift_rotor_hub_weight   = 4.   * Units.kg
            prop_hub_weight         = 4.   * Units.kg
            lift_rotor_BRS_weight   = 16.  * Units.kg
    
            # Rotor Weight
            number_of_propellers    = 0.0  
            number_of_lift_rotors   = 0.0   
            total_number_of_rotors  = 0.0      
            lift_rotor_servo_weight = 0.0  
            
        for propulsor in network.propulsors:
            rotor = propulsor.rotor   
            if type(rotor) == RCAIDE.Library.Components.Propulsors.Converters.Propeller:
                ''' Propeller Weight '''  
                number_of_propellers       += 1   
                rTip_ref                   = rotor.tip_radius 
                bladeSol_ref               = rotor.blade_solidity 
                prop_servo_weight          = 5.2 * Units.kg  
                propeller_mass             = compute_rotor_weight(rotor, maxLift/5.) * Units.kg
                weight.rotors              += propeller_mass 
                rotor.mass_properties.mass  =  propeller_mass + prop_hub_weight + prop_servo_weight
                maxVTip                     = rotor.cruise.design_angular_velocity * rotor.tip_radius
                weight.servos              += prop_servo_weight
                weight.hubs                += prop_hub_weight
                
            if (type(rotor) == RCAIDE.Library.Components.Propulsors.Converters.Lift_Rotor or type(rotor) == RCAIDE.Library.Components.Propulsors.Converters.Prop_Rotor) or type(rotor) == RCAIDE.Library.Components.Propulsors.Converters.Rotor:
                ''' Lift Rotor, Prop-Rotor or Rotor Weight '''  
                number_of_lift_rotors       += 1  
                rTip_ref                    = rotor.tip_radius
                bladeSol_ref                = rotor.blade_solidity 
                maxVTip                     = rotor.hover.design_angular_velocity * rotor.tip_radius
                lift_rotor_servo_weight     = 0.65 * Units.kg 
                if rotor.oei.design_thrust == None:
                    design_thrust = rotor.hover.design_thrust
                else:
                    design_thrust =rotor.oei.design_thrust
                lift_rotor_mass             = compute_rotor_weight(rotor,design_thrust)
                weight.rotors               += lift_rotor_mass 
                rotor.mass_properties.mass  =  lift_rotor_mass + lift_rotor_hub_weight + lift_rotor_servo_weight
                weight.servos               += lift_rotor_servo_weight
                weight.hubs                 += lift_rotor_hub_weight 
            
            # Motor 
            eta             = propulsor.motor.efficiency  
            weight.motors  += propulsor.motor.mass_properties.mass  
                
        total_number_of_rotors  = int(number_of_lift_rotors + number_of_propellers)  
        if total_number_of_rotors > 1:
            prop_BRS_weight     = 16.   * Units.kg
        else:
            prop_BRS_weight     = 0.   * Units.kg
 
        # Add associated weights  
        weight.BRS   += (prop_BRS_weight + lift_rotor_BRS_weight)  
        maxLiftPower = 1.15*maxLift*(disk_area_factor*np.sqrt(maxLift/(2*rho_ref*np.pi*rTip_ref**2)) + bladeSol_ref*AvgBladeCD/8*maxVTip**3/(maxLift/(rho_ref*np.pi*rTip_ref**2)))
        # Tail Rotor
        if number_of_lift_rotors == 1: # this assumes that the vehicle is an electric helicopter with a tail rotor 
            maxLiftOmega   = maxVTip/rTip_ref
            maxLiftTorque  = maxLiftPower / maxLiftOmega  
            for bus in network.busses: 
                tailrotor = next(iter(bus.lift_rotors))
                weight.tail_rotor  = compute_rotor_weight(tailrotor, 1.5*maxLiftTorque/(1.25*rTip_ref))*0.2 * Units.kg
                weight.rotors     += weight.tail_rotor 

    #-------------------------------------------------------------------------------
    # Thermal Management System Weight
    #-------------------------------------------------------------------------------
    tms_weight = 0.0
    for network in vehicle.networks:
        for coolant_line in network.coolant_lines:
            weight.thermal_management_system.battery_module = Data()  # Add container for battery module
            for i, battery_module in enumerate(coolant_line.battery_modules):
                module_key = f'module_{i+1}'  # Create unique key for each module
                weight.thermal_management_system.battery_module[module_key] = 0.0  # Initialize weight
                for HAS in battery_module:
                    weight.thermal_management_system.battery_module[module_key] = HAS.mass_properties.mass
                    tms_weight +=  HAS.mass_properties.mass

            for tag, item in coolant_line.items(): 
                if tag == 'heat_exchangers':
                    for heat_exchanger in item:
                        weight.thermal_management_system[heat_exchanger.tag] = heat_exchanger.mass_properties.mass
                        tms_weight +=  heat_exchanger.mass_properties.mass
                if tag == 'reservoirs':
                    for reservoir in item:
                        weight.thermal_management_system[reservoir.tag] = reservoir.mass_properties.mass
                        tms_weight +=  reservoir.mass_properties.mass
    weight.thermal_management_system.total = tms_weight
    #-------------------------------------------------------------------------------
    # Wing and Motor Wiring Weight
    #-------------------------------------------------------------------------------
    maxSpan =  0
    for wing in vehicle.wings: 
        maxSpan = max(wing.spans.projected, maxSpan)  
        if wing.symbolic:
            wing_weight = 0
        else:
            wing_weight               = compute_wing_weight(wing, vehicle, maxLift/5, safety_factor= safety_factor, max_g_load =  max_g_load )
            wing_tag                  = wing.tag
            weight.wings[wing_tag]    = wing_weight
            wing.mass_properties.mass = wing_weight 
        weight.wings_total           += wing_weight

        # compute_wiring_weight weight
        if isinstance(wing, RCAIDE.Library.Components.Wings.Main_Wing):
            wiring_weight  = compute_wiring_weight(wing, vehicle, maxLiftPower/(eta*total_number_of_rotors)) * Units.kg
        else:
            wiring_weight =  0
        weight.wiring  += wiring_weight 

    #-------------------------------------------------------------------------------
    # Landing Gear Weight
    #-------------------------------------------------------------------------------  
    main_landing_gear =  False
    for LG in vehicle.landing_gears:
        if isinstance(LG, RCAIDE.Library.Components.Landing_Gear.Main_Landing_Gear):
            LG.mass_properties.mass =  weight.landing_gear
            main_landing_gear = True 
    if main_landing_gear == False:
        main_gear = RCAIDE.Library.Components.Landing_Gear.Main_Landing_Gear()  
        main_gear.mass_properties.mass =  weight.landing_gear
        vehicle.append_component(main_gear)
        
    #-------------------------------------------------------------------------------
    # Fuselage  Weight  
    #-------------------------------------------------------------------------------
    for fuse in  vehicle.fuselages: 
        fuselage_weight = compute_fuselage_weight(fuse, maxSpan, MTOW )  
        fuse.mass_properties.center_of_gravity[0][0] = .45*fuse.lengths.total
        fuse.mass_properties.mass                    =  fuselage_weight + weight.passengers + weight.seats + weight.wiring + weight.BRS
        weight.fuselage += fuselage_weight  

    #-------------------------------------------------------------------------------
    # Boom Weight
    #-------------------------------------------------------------------------------
    for boom in vehicle.booms:
        boom_weight                = compute_boom_weight(boom) * Units.kg
        weight.booms               += boom_weight 
        boom.mass_properties.mass  =  boom_weight 
    
    #-------------------------------------------------------------------------------
    # Pack Up Outputs
    #-------------------------------------------------------------------------------
    output                                            = Data()
    output.empty                                      = Data()    
    output.empty.structural                           = Data()
    output.empty.structural.booms                     = miscelleneous_weight_factor *weight.booms 
    output.empty.structural.fusleage                  = miscelleneous_weight_factor *weight.fuselage  
    output.empty.structural.landing_gear              = miscelleneous_weight_factor *weight.landing_gear  
    output.empty.structural.wings                     = miscelleneous_weight_factor *weight.wings_total 
    output.empty.structural.total                     = weight.booms + weight.fuselage + weight.landing_gear +weight.wings_total 
             
    output.empty.propulsion                           = Data()
    output.empty.propulsion.motors                    = miscelleneous_weight_factor *weight.motors
    output.empty.propulsion.rotors                    = miscelleneous_weight_factor *weight.rotors
    output.empty.propulsion.hubs                      = miscelleneous_weight_factor *weight.hubs 
    output.empty.propulsion.servos                    = miscelleneous_weight_factor *weight.servos
    output.empty.propulsion.wiring                    = miscelleneous_weight_factor *weight.wiring 
    output.empty.propulsion.battery                   = miscelleneous_weight_factor *weight.battery
    output.empty.propulsion.TMS                       = miscelleneous_weight_factor *weight.thermal_management_system.total
    output.empty.propulsion.total                     = weight.rotors + weight.hubs +  weight.battery +  weight.motors +   weight.wiring +   weight.servos +  weight.thermal_management_system.total 

    output.empty.systems                              = Data()
    output.empty.systems.environmental_control_system = miscelleneous_weight_factor * weight.ECS
    output.empty.systems.avionics                     = miscelleneous_weight_factor * weight.avionics  
    output.empty.systems.seats                        = miscelleneous_weight_factor * weight.seats  
    output.empty.systems.balistic_recovery_system     = miscelleneous_weight_factor * weight.BRS
    output.empty.systems.total                        = weight.ECS + weight.avionics +   weight.BRS +  weight.seats  
    
    output.empty.total        = output.empty.systems.total +  output.empty.propulsion.total +  output.empty.structural.total
    output.payload            = Data()
    output.payload.total      = weight.passengers + weight.payload
    output.payload.passengers = weight.passengers  
    output.payload.payload    = weight.payload 
    output.zero_fuel_weight   = output.empty.total + output.payload.total 
    output.fuel               = 0
    output.total              = output.empty.total + output.payload.total 
    return output









# RCAIDE/Library/Methods/Propulsors/Converters/Ducted_Fan.py
#  
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports
import  RCAIDE
from RCAIDE.Framework.Analyses.Propulsion.Ducted_Fan_Design_Code import Ducted_Fan_Design_Code
from RCAIDE.Framework.Core import Data ,redirect   
from RCAIDE.Library.Plots  import *
 
# python imports   
from shutil import rmtree
from .write_geometry                     import  write_geometry
from .write_input_deck                   import  write_input_deck
from .run_dfdc_analysis                  import  run_dfdc_analysis
from .translate_conditions_to_dfdc_cases import  translate_conditions_to_dfdc_cases
from .read_results                       import  read_results 
from scipy import interpolate 
import os
import numpy as  np

# ----------------------------------------------------------------------------------------------------------------------
#  design_ducted_fan
# ---------------------------------------------------------------------------------------------------------------------- 
def design_ducted_fan(ducted_fan, dfdc_bin_name = 'dfdc', new_regression_results = False, keep_files = True): 
    """ Optimizes ducted fan given input design conditions.

    Assumptions: 

    Source:
        https://web.mit.edu/drela/Public/web/dfdc/
        http://www.esotec.org/sw/DFDC.html 
    
    Inputs:
        dfdc_analysis (dict): DFDC analysis data structure  

    Outputs:
        None
    """

    if ducted_fan.cruise.design_altitude == None:
        raise AttributeError('design altitude not set')
    atmosphere      = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmo_data       = atmosphere.compute_values(ducted_fan.cruise.design_altitude)      
    
    if ducted_fan.cruise.design_freestream_mach == None: 
        ducted_fan.cruise.design_freestream_mach = ducted_fan.cruise.design_freestream_velocity / atmo_data.speed_of_sound[0,0]     
    if ducted_fan.cruise.design_reference_mach == None:  
        ducted_fan.cruise.design_reference_mach = ducted_fan.cruise.design_reference_velocity / atmo_data.speed_of_sound[0,0] 

    if ducted_fan.cruise.design_freestream_velocity == None: 
        ducted_fan.cruise.design_freestream_velocity = ducted_fan.cruise.design_freestream_mach * atmo_data.speed_of_sound[0,0]     
    if  ducted_fan.cruise.design_reference_velocity== None:  
        ducted_fan.cruise.design_reference_velocity = ducted_fan.cruise.design_reference_mach * atmo_data.speed_of_sound[0,0]        
    
    if (ducted_fan.cruise.design_angular_velocity)  == None   and (ducted_fan.cruise.design_tip_mach == None):
        raise AttributeError('design tip mach and/or angular velocity not set') 
    if ducted_fan.cruise.design_angular_velocity  == None:  
        ducted_fan.cruise.design_angular_velocity    = (ducted_fan.cruise.design_tip_mach *atmo_data.speed_of_sound[0,0]) /ducted_fan.tip_radius          
    if ducted_fan.cruise.design_tip_mach == None:
        ducted_fan.cruise.design_tip_mach  =  (ducted_fan.cruise.design_angular_velocity * ducted_fan.tip_radius) *atmo_data.speed_of_sound[0,0]
        
    dfdc_analysis                                   = Ducted_Fan_Design_Code() 
    dfdc_analysis.geometry                          = ducted_fan
    dfdc_analysis.settings.filenames.dfdc_bin_name  = dfdc_bin_name
    dfdc_analysis.settings.new_regression_results   = new_regression_results
    dfdc_analysis.settings.keep_files               = keep_files  
    run_folder                                      = os.path.abspath(dfdc_analysis.settings.filenames.run_folder)
    run_script_path                                 = run_folder.rstrip('dfdc_files').rstrip('/')    
    deck_template                                   = dfdc_analysis.settings.filenames.deck_template 
    print_output                                    = dfdc_analysis.settings.print_output  
    dfdc_analysis.current_status.deck_file          = deck_template.format(1) 
 
    # translate conditions  
    translate_conditions_to_dfdc_cases(dfdc_analysis)  
    dfdc_analysis.settings.filenames.case  =  dfdc_analysis.geometry.tag +  '.case'
    # write the input files
    with redirect.folder(run_folder,force=False):
        write_geometry(dfdc_analysis,run_script_path)    
        write_input_deck(dfdc_analysis)   

        # RUN DFDC!
        _ = run_dfdc_analysis(dfdc_analysis,print_output)

    # translate results
    results = read_results(dfdc_analysis)

    if not dfdc_analysis.settings.keep_files:
        rmtree(run_folder) 
         
    # save blade geometry 
    ducted_fan.rotor.twist_distribution           = results.geometry.rotor_twist_distribution            
    ducted_fan.rotor.chord_distribution           = results.geometry.rotor_chord_distribution            
    ducted_fan.rotor.radius_distribution          = results.geometry.rotor_radius_distribution           
    ducted_fan.rotor.non_dim_radius_distribution  = results.geometry.rotor_non_dim_radius_distribution   
    ducted_fan.rotor.solidity_distribution        = results.geometry.rotor_solidity_distribution         
    ducted_fan.stator.twist_distribution          = results.geometry.stator_twist_distribution           
    ducted_fan.stator.chord_distribution          = results.geometry.stator_chord_distribution           
    ducted_fan.stator.radius_distribution         = results.geometry.stator_radius_distribution          
    ducted_fan.stator.non_dim_radius_distribution = results.geometry.stator_non_dim_radius_distribution  
    ducted_fan.stator.solidity_distribution       = results.geometry.stator_solidity_distribution   

    mach              = dfdc_analysis.training.mach       
    tip_mach          = dfdc_analysis.training.tip_mach         
    altitude          = dfdc_analysis.training.altitude/1000 
    
    ducted_fan.cruise.design_thrust             = results.performance.design_thrust            
    ducted_fan.cruise.design_power              = results.performance.design_power             
    ducted_fan.cruise.design_efficiency         = results.performance.design_efficiency        
    ducted_fan.cruise.design_torque             = results.performance.design_power /  ducted_fan.cruise.design_angular_velocity   
    ducted_fan.cruise.design_thrust_coefficient = results.performance.design_thrust_coefficient
    ducted_fan.cruise.design_power_coefficient  = results.performance.design_power_coefficient  
    
    # create performance surrogates 
    raw_data           = results.performance
    thrust             = clean_data(raw_data.thrust,mach,tip_mach,altitude,raw_data.converged_solution)               
    power              = clean_data(raw_data.power,mach,tip_mach,altitude,raw_data.converged_solution)                
    efficiency         = clean_data(raw_data.efficiency,mach,tip_mach,altitude,raw_data.converged_solution)           
    torque             = clean_data(raw_data.torque,mach,tip_mach,altitude,raw_data.converged_solution)               
    thrust_coefficient = clean_data(raw_data.thrust_coefficient,mach,tip_mach,altitude,raw_data.converged_solution)   
    power_coefficient  = clean_data(raw_data.power_coefficient,mach,tip_mach,altitude,raw_data.converged_solution)     
    
    surrogates =  Data()
    surrogates.thrust              = interpolate.RegularGridInterpolator((mach,tip_mach,altitude),thrust               ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.power               = interpolate.RegularGridInterpolator((mach,tip_mach,altitude),power                ,method = 'linear',   bounds_error=False, fill_value=None)
    surrogates.efficiency          = interpolate.RegularGridInterpolator((mach,tip_mach,altitude),efficiency           ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.torque              = interpolate.RegularGridInterpolator((mach,tip_mach,altitude),torque               ,method = 'linear',   bounds_error=False, fill_value=None)
    surrogates.thrust_coefficient  = interpolate.RegularGridInterpolator((mach,tip_mach,altitude),thrust_coefficient   ,method = 'linear',   bounds_error=False, fill_value=None)      
    surrogates.power_coefficient   = interpolate.RegularGridInterpolator((mach,tip_mach,altitude),power_coefficient    ,method = 'linear',   bounds_error=False, fill_value=None) 
    
    ducted_fan.performance_surrogates =  surrogates
    return

def clean_data(raw_data,mach,tip_mach,altitude,convergence_matrix):  
    cleaned_data =  np.zeros((len(mach),len(tip_mach),len(altitude))) 
    for i in range(len(mach)): 
        x = tip_mach
        y = altitude
        
        # mask invalid values
        array = np.ma.masked_invalid(raw_data[i])
        if np.all(array.mask ==False):
            cleaned_data[i, :, :] =  array.data
        else: 
            yy,xx   = np.meshgrid(y, x)
            x1      = xx[~array.mask]
            y1      = yy[~array.mask]
            newarr  = array[~array.mask] 
            points  = np.vstack((x1, y1)).T
            values1 = newarr.data
            cleaned_data[i, :, :]  = interpolate.griddata(points, values1, (xx, yy), method='cubic',  fill_value = -1E5)
             
    return cleaned_data




# RCAIDE/Library/Components/Propulsors/Converters/Ram.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Feb 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports
from RCAIDE.Framework.Core import Data
from RCAIDE.Library.Components                      import Component  
from RCAIDE.Library.Methods.Propulsors.Converters.Ram.append_ram_conditions import append_ram_conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  Fan Component
# ----------------------------------------------------------------------------------------------------------------------
class Ram(Component):
    """
    A ram air compression component model that simulates the compression of incoming air flow.

    Attributes
    ----------
    tag : str
        Identifier for the ram component. Default is 'Ram'.
        
    working_fluid : Data
        Container for working fluid properties and conditions. Default is empty Data().

    Notes
    -----
    The Ram class models the compression of incoming air due to the ram effect
    in high-speed flight. It calculates:

    * Total pressure rise due to flow deceleration
    * Temperature increase from compression
    * Changes in fluid properties
    * Ram recovery efficiency
    * Compressibility effects

    The model assumes:

    * Quasi-one-dimensional flow
    * Adiabatic process (no heat transfer)
    * Perfect gas behavior
    * Steady flow conditions
    * No boundary layer effects
    * No shock waves (unless explicitly modeled)

    **Theory**
    The ram compression process follows isentropic flow relations for subsonic flow
    and includes normal shock relations for supersonic flow. The total pressure 
    recovery depends on the flight Mach number and inlet geometry.

    **Definitions**
    
    'Ram Effect'
        Pressure rise due to deceleration of high-speed flow
    'Recovery Factor'
        Ratio of actual to ideal pressure rise
    'Working Fluid'
        Fluid medium undergoing compression (typically air)

    References
    ----------
    [1] Mattingly, J. D., & Boyer, K. M. (2016). Elements of propulsion: Gas 
        turbines and rockets, second edition Jack D. Mattingly, Keith M. Boyer. 
        American Institute of Aeronautics and Astronautics.

    See Also
    --------
    RCAIDE.Library.Components.Component
    RCAIDE.Library.Methods.Propulsors.Converters.Ram.append_ram_conditions
    """

    def __defaults__(self):
        """This sets the default values for the component to function.

        Assumptions:
            None

        Source:
        [1] Mattingly, J. D., & Boyer, K. M. (2016). Elements of propulsion: Gas 
            turbines and rockets, second edition Jack D. Mattingly, Keith M. Boyer. 
            American Institute of Aeronautics and Astronautics.
        """
        #set the deafult values
        self.tag                      = 'Ram' 
        self.working_fluid            = Data()


    def append_operating_conditions(self,segment,propulsor): 
        propulsor_conditions =  segment.state.conditions.energy[propulsor.tag]
        append_ram_conditions(self,segment,propulsor_conditions)
        return                         


## @ingroup Methods-Aerodynamics-Common-Fidelity_Zero-Lift
# extract_wing_VD.py
#
# Created:   Feb 2022, R. Erhard
# Modified:  Mar 2022, R. Erhard
#            Apr 2022, E. Botero

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------
# RCAIDE imports
import  RCAIDE
from RCAIDE.Framework.Core import Data 

# package imports
import numpy as np 


## @ingroup Methods-Aerodynamics-Common-Fidelity_Zero-Lift
def extract_wing_collocation_points(geometry, wing_instance_idx):

    """ This extracts the collocation points of the vehicle vortex distribution
    belonging to the specified wing instance index. This is useful for slipstream
    analysis, where the wake of a propeller is included in the VLM analysis 
    of a specified wing in the vehicle.

    Source:
    None

    Inputs:
    geometry      -    SUAVE vehicle 
    wing_instance -    wing instance tag

    Outputs:
    VD_wing   - colocation points of vortex distribution for specified wing
    ids       - indices in the vortex distribution corresponding to specified wing

    Properties Used:
    N/A
    """
    # unpack vortex distribution properties
    try:
        VD           = geometry.vortex_distribution
    except:
        print("No vortex distribution defined. Creating default vortex distribution from vehicle.")
        settings = RCAIDE.Framework.Analyses.Aerodynamics.Vortex_Lattice_Method().settings
        VD = RCAIDE.Library.Methods.Aerodynamics.Vortex_Lattice_Method.generate_vortex_distribution(geometry,settings)        
        
    sym          = VD.symmetric_wings
    
    # Find the beginning and end indices of the wing
    # all breaks
    breaks = np.hstack([0,np.cumsum(VD.n_cw*VD.n_sw)])
    
    # Find the initial index of the wing
    semispan_idx     = wing_instance_idx + np.sum(sym[0:wing_instance_idx])
    start_pt         = breaks[semispan_idx]
    
    # Find the final index of the wing
    end_semispan_idx = semispan_idx + 1 + sym[wing_instance_idx]
    end_pt              = breaks[end_semispan_idx]
    
    # Make ranges of points
    ids              = range(semispan_idx, end_semispan_idx)
    pt_ids           = range(start_pt,end_pt)

    # Pack the wing level resultss
    VD_wing  = Data()
    VD_wing.XC   = VD.XC[pt_ids]
    VD_wing.YC   = VD.YC[pt_ids]
    VD_wing.ZC   = VD.ZC[pt_ids]
    
    VD_wing.XA1  = VD.XA1[pt_ids]
    VD_wing.XA2  = VD.XA2[pt_ids]
    VD_wing.XB1  = VD.XB1[pt_ids]
    VD_wing.XB2  = VD.XB2[pt_ids]
    VD_wing.YA1  = VD.YA1[pt_ids]
    VD_wing.YA2  = VD.YA2[pt_ids]
    VD_wing.YB1  = VD.YB1[pt_ids]
    VD_wing.YB2  = VD.YB2[pt_ids]
    VD_wing.ZA1  = VD.ZA1[pt_ids]
    VD_wing.ZA2  = VD.ZA2[pt_ids]
    VD_wing.ZB1  = VD.ZB1[pt_ids]
    VD_wing.ZB2  = VD.ZB2[pt_ids]    
    
    
    VD_wing.n_cp = len(VD_wing.XC)
    VD_wing.n_cw = VD.n_cw[ids]
    VD_wing.n_sw = VD.n_sw[ids]

    return VD_wing, pt_ids





# RCAIDE/Library/Methods/Thermal_Management/Batteries/Liquid_Cooled_Wavy_Channel/design_wavy_channel

# Created:  Apr 2024, S. Shekar 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports   
from RCAIDE.Framework.Core                                                                                         import Units , Data  
from RCAIDE.Library.Methods.Thermal_Management.Batteries.Liquid_Cooled_Wavy_Channel.wavy_channel_sizing_setup      import wavy_channel_sizing_setup
from RCAIDE.Library.Methods.Thermal_Management.Batteries.Liquid_Cooled_Wavy_Channel.wavy_channel_geometry_setup    import wavy_channel_geometry_setup
from RCAIDE.Framework.Optimization.Common                                                                          import Nexus
from RCAIDE.Framework.Optimization.Packages.scipy                                                                  import scipy_setup

# Python package imports  
import numpy as np  
import time 

# ----------------------------------------------------------------------------------------------------------------------
#  Wavy Channel Design
# ----------------------------------------------------------------------------------------------------------------------
def design_wavy_channel(HAS,battery,single_side_contact=True, dry_mass=True,
                        solver_name= 'SLSQP',iterations = 200,solver_sense_step = 1E-4,
                                       solver_tolerance = 1E-3,print_iterations = False):  
    
    """ Optimizes wavy channel geometric properties input parameters to minimize either design power and mass. 
        This scrip adopts RCAIDE's native optimization style where the objective function is expressed 
        as a sizing function, considering both power and mass.
          
          Inputs: 
          HAS.
              Maximum heat generated by the battery pack. 
              Wavy Channel width
              Wavy Channel thhickness
              Wavy  Channel contact angle
             
          Outputs:
                 Wavy Channel width
                 Wavy Channel thhickness
                 Wavy  Channel contact angle
                 Mass of Heat Acquisition System
                 Power Consumed by Heat Acquisition System 
                 
          Assumptions: 
             The wavy channel extracts from the battery pack considering it to be a lumped mass.  
        
          Source:
            Zhao, C., Clarke, M., Kellermann H., Verstraete D., “Design of a Liquid Cooling System for Lithium-Ion Battery Packs for eVTOL Aircraft" 
    """    
    
    if HAS.coolant_inlet_temperature == None:
        assert('specify coolant inlet temperature')
    elif HAS.design_battery_operating_temperature  == None:
        assert('specify design battery temperature')  
    elif HAS.design_heat_removed  == None: 
        assert('specify heat generated') 

    # start optimization 
    ti                   = time.time()   
    optimization_problem = wavy_channel_design_problem_setup(HAS,battery,print_iterations)
    output               = scipy_setup.SciPy_Solve(optimization_problem,solver=solver_name, iter = iterations , sense_step = solver_sense_step,tolerance = solver_tolerance)  

    tf                   = time.time()
    elapsed_time         = round((tf-ti)/60,2)
    print('Channel Cooling hex Optimization Simulation Time: ' + str(elapsed_time) + ' mins')   

    # print optimization results 
    print (output)   
    HAS_opt = optimization_problem.hrs_configurations.optimized.networks.electric.coolant_lines.coolant_line.battery_modules[battery.tag].thermal_management_system.heat_acquisition_system
    HAS.mass_properties.mass       = HAS_opt.mass_properties.mass      
    HAS.design_power_draw          = HAS_opt.design_power_draw         
    HAS.design_heat_removed        = HAS_opt.design_heat_removed       
    HAS.coolant_outlet_temperature = HAS_opt.coolant_outlet_temperature
    HAS.coolant_pressure_drop      = HAS_opt.coolant_pressure_drop
    HAS.channel_side_thickness     = HAS_opt.channel_side_thickness     
    HAS.channel_width              = HAS_opt.channel_width          
    HAS.coolant_flow_rate          = HAS_opt.coolant_flow_rate  # per module
    HAS.channel_contact_angle      = HAS_opt.channel_contact_angle  
    HAS.channel_area               = HAS_opt.surface_area_channel 
    
    # Update Battery Spacing based on Wavy Channel
    battery.geometrtic_configuration.parallel_spacing  = HAS_opt.battery_parllel_spacing 
    battery.geometrtic_configuration.normal_spacing    = HAS_opt.battery_series_spacing 

    return HAS

def wavy_channel_design_problem_setup(HAS,battery,print_iterations):
    """ Sets up wavy channel optimization problem including design variables, constraints and objective function
        using RCAIDE's Nexus optimization framework. Appends methodolody of planform modification to Nexus.
          Inputs: 
             HAS     - heat acqusition data structure           [None]
             
          Outputs: 
              nexus    - RCAIDE's optimization framework [None]
              
          Assumptions:
            None
        
          Source:
             None
    """        

    nexus                        = Nexus()
    problem                      = Data()
    nexus.optimization_problem   = problem 

    b_0        = HAS.channel_side_thickness                           
    d_0        = HAS.channel_width          
    m_dot_0    = HAS.coolant_flow_rate  
    theta_0    = HAS.channel_contact_angle       

    # ---------------------------------------------------------------------------------------------------------- 
    # Design Variables 
    # ----------------------------------------------------------------------------------------------------------       
    inputs = []   
    #               variable   initial      lower limit           upper limit       scaling       units 
    inputs.append([ 'm_dot' ,  m_dot_0     ,   0.05         ,  2.5                , 1E-1       ,  1*Units.less])   
    inputs.append([ 'd'     ,  d_0         ,   0.001        ,  0.01               , 1E-3       ,  1*Units.less])  
    inputs.append([ 'b'     ,  b_0         ,   0.001        ,  0.002              , 1E-3       ,  1*Units.less]) 
    inputs.append([ 'theta' ,  theta_0     ,48*Units.degrees, 70*Units.degrees    , 1.0        ,  1*Units.less])         
    problem.inputs = np.array(inputs,dtype=object)    

    # ----------------------------------------------------------------------------------------------------------
    # Objective
    # ---------------------------------------------------------------------------------------------------------- 
    problem.objective = np.array([  
        [  'Obj'  ,  10   ,    1*Units.less] ],dtype=object)


    # ----------------------------------------------------------------------------------------------------------
    # Constraints
    # ----------------------------------------------------------------------------------------------------------  
    constraints = []    
    constraints.append([ 'Q_con'         ,  '>'  ,  0.0 ,   1.0   , 1*Units.less]) 
    constraints.append([ 'thick_con'     ,  '>'  ,  0.0 ,   1.0   , 1*Units.less]) 
    
    problem.constraints =  np.array(constraints,dtype=object)                

    # ----------------------------------------------------------------------------------------------------------
    #  Aliases
    # ---------------------------------------------------------------------------------------------------------- 
    aliases = [] 
    btms = 'hrs_configurations.optimized.networks.electric.coolant_lines.coolant_line.battery_modules.'+battery.tag+'.thermal_management_system.heat_acquisition_system'  
    aliases.append([ 'm_dot'       , btms + '.coolant_flow_rate'])    
    aliases.append([ 'b'           , btms + '.channel_side_thickness']) 
    aliases.append([ 'd'           , btms + '.channel_width']) 
    aliases.append([ 'theta'       , btms + '.channel_contact_angle']) 
    aliases.append([ 'Obj'         , 'summary.mass_power_objective' ])  
    aliases.append([ 'Q_con'       , 'summary.heat_energy_constraint'])  
    aliases.append([ 'thick_con'   , 'summary.thickness_constraint'])
    problem.aliases = aliases

    # -------------------------------------------------------------------
    #  Vehicles
    # ------------------------------------------------------------------- 
    nexus.hrs_configurations = wavy_channel_geometry_setup(HAS,battery)

    # -------------------------------------------------------------------
    #  Analyses
    # -------------------------------------------------------------------
    nexus.analyses = None 

    # -------------------------------------------------------------------
    #  Missions
    # -------------------------------------------------------------------
    nexus.missions = None

    # -------------------------------------------------------------------
    #  Procedure
    # -------------------------------------------------------------------    
    #nexus.print_iterations  = print_iterations 
    nexus.procedure         = wavy_channel_sizing_setup()

    # -------------------------------------------------------------------
    #  Summary
    # -------------------------------------------------------------------    
    nexus.summary         = Data()     

    return nexus


# RCAIDE/Library/Methods/Propulsors/Converters/Compressor/compute_compressor_performance.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke

# ---------------------------------------------------------------------------------------------------------------------- 
# Imports 
# ----------------------------------------------------------------------------------------------------------------------
import numpy as np

# ---------------------------------------------------------------------------------------------------------------------- 
# compute_compression_nozzle_performance
# ----------------------------------------------------------------------------------------------------------------------
def compute_compressor_performance(compressor,compressor_conditions,conditions):
    """ Computes the performance of a compressor bases on its polytropic efficiency.
        The following properties are computed: 
       compressor.outputs.
         stagnation_temperature  (numpy.ndarray): exit stagnation_temperature   [K]  
         stagnation_pressure     (numpy.ndarray): exit stagnation_pressure      [Pa]
         stagnation_enthalpy     (numpy.ndarray): exit stagnation_enthalpy      [J/kg]
         work_done               (numpy.ndarray): work done                     [J/kg] 

    Assumptions:
        Constant polytropic efficiency and pressure ratio

    Source:
        https://web.stanford.edu/~cantwell/AA283_Course_Material/AA283_Course_Notes/

    Args:
        conditions.freestream.
          isentropic_expansion_factor         (numpy.ndarray): isentropic_expansion_factor        [unitless]
          specific_heat_at_constant_pressure  (numpy.ndarray): specific_heat_at_constant_pressure [J/(kg K)]
        compressor.
           inputs.stagnation_temperature      (numpy.ndarray): entering stagnation temperature [K]
           inputs.stagnation_pressure         (numpy.ndarray): entering stagnation pressure    [Pa] 
           pressure_ratio                             (float): pressure ratio                  [unitless]
           polytropic_efficiency                      (float): polytropic efficiency           [unitless]

    Returns:
        None 
    """          
    
    # Unpack component inputs
    Tt_in    = compressor_conditions.inputs.stagnation_temperature
    Pt_in    = compressor_conditions.inputs.stagnation_pressure 
    PR       = compressor.pressure_ratio
    etapold  = compressor.polytropic_efficiency 
    T0       = compressor_conditions.inputs.static_temperature
    P0       = compressor_conditions.inputs.static_pressure  
    M0       = compressor_conditions.inputs.mach_number    
    
    # Unpack ram inputs
    working_fluid           = compressor.working_fluid
 
    # Compute the working fluid properties 
    gamma  = working_fluid.compute_gamma(T0,P0) 
    Cp     = working_fluid.compute_cp(T0,P0)    
    R      = working_fluid.compute_R(T0,P0)
        
    # Compute the output properties based on the pressure ratio of the component
    ht_in     = Tt_in*Cp 
    Pt_out    = Pt_in*PR
    Tt_out    = Tt_in*(PR**((gamma-1)/(gamma*etapold)))
    ht_out    = Tt_out*Cp
    T_out     = Tt_out/(1.+(gamma-1.)/2.*M0*M0)
    P_out     = Pt_out/((1.+(gamma-1.)/2.*M0*M0)**(gamma/(gamma-1.))) 
    M_out     = np.sqrt( (((Pt_out/P_out)**((gamma-1.)/gamma))-1.) *2./(gamma-1.) ) 
    # Compute the work done by the compressor (normalized by mass flow i.e. J/(kg/s)
    work_done = ht_out - ht_in
    
    # Pack results  
    compressor_conditions.outputs.work_done               = work_done 
    compressor_conditions.outputs.stagnation_temperature  = Tt_out
    compressor_conditions.outputs.stagnation_pressure     = Pt_out
    compressor_conditions.outputs.stagnation_enthalpy     = ht_out
    compressor_conditions.outputs.static_temperature      = T_out
    compressor_conditions.outputs.static_pressure         = P_out 
    compressor_conditions.outputs.mach_number             = M_out
    compressor_conditions.outputs.gas_constant            = R
    compressor_conditions.outputs.gamma                   = gamma 
    compressor_conditions.outputs.cp                      = Cp   
    
    return 




# RCAIDE/Library/Missions/Segments/Untrimmed/Untrimmed.py
# 
# 
# Created:  Jul 2023, M. Clarke 
 
# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------  

import numpy as np

# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------   
def initialize_conditions(segment):
    """Sets the specified conditions which are given for the segment type.

    Assumptions:
    A fixed speed and altitude

    Source:
    N/A

    Inputs:
    segment.altitude                               [meters]
    segment.air_speed                              [meters/second]
    segment.linear_acceleration_x                         [meters/second^2]
    segment.sideslip_angle                         [radians]
    segment.linear_acceleration_z                         [meters/second^2]

    Outputs:
    conditions.frames.inertial.acceleration_vector [meters/second^2]
    conditions.frames.inertial.velocity_vector     [meters/second]
    conditions.frames.inertial.position_vector     [meters]
    conditions.freestream.altitude                 [meters]
    conditions.frames.inertial.time                [seconds]

    Properties Used:
    N/A
    """      
    
    # unpack
    alt                    = segment.altitude
    air_speed              = segment.air_speed  
    sideslip               = segment.sideslip_angle
    linear_acceleration_x  = segment.linear_acceleration_x
    linear_acceleration_y  = segment.linear_acceleration_y 
    linear_acceleration_z  = segment.linear_acceleration_z
    roll_rate              = segment.roll_rate
    pitch_rate             = segment.pitch_rate
    yaw_rate               = segment.yaw_rate
    
    
    # check for initial altitude
    if alt is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt = -1.0 *segment.state.initials.conditions.frames.inertial.position_vector[-1,2]
    
    # pack
    air_speed_x                                                   = np.cos(sideslip)*air_speed 
    air_speed_y                                                   = np.sin(sideslip)*air_speed 
    segment.state.conditions.freestream.altitude[:,0]             = alt
    segment.state.conditions.frames.inertial.position_vector[:,2] = -alt # z points down
    segment.state.conditions.frames.inertial.velocity_vector[:,0] = air_speed_x
    segment.state.conditions.frames.inertial.velocity_vector[:,1] = air_speed_y
    segment.state.conditions.frames.inertial.acceleration_vector  = np.array([[linear_acceleration_x,linear_acceleration_y,linear_acceleration_z]])  
    segment.state.conditions.static_stability.roll_rate[:,0]      = roll_rate         
    segment.state.conditions.static_stability.pitch_rate[:,0]     = pitch_rate
    segment.state.conditions.static_stability.yaw_rate[:,0]       = yaw_rate      


# RCAIDE/Energy/Thermal_Management/Batteries/__init__.py

"""
Module containing thermal management components specifically for battery systems. This module 
provides classes for different battery cooling approaches including air cooling, cryogenic 
cooling, and liquid cooling with wavy channels.

See Also
--------
RCAIDE.Library.Components.Energy.Sources.Battery_Modules
    Related module for battery system modeling
RCAIDE.Library.Components.Thermal_Management.Heat_Exchangers
    Related module for heat exchanger components used in battery cooling
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
from .Air_Cooled                 import Air_Cooled
from .Cryocooler                 import Cryocooler
from .Liquid_Cooled_Wavy_Channel import Liquid_Cooled_Wavy_Channel


# RCAIDE/Library/Methods/Propulsors/Converters/Fan/compute_fan_performance.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Feb 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------    
# package imports
import numpy as np 
from RCAIDE.Library.Methods.Gas_Dynamics.fm_id import fm_id

# exceptions/warnings
from warnings import warn

# ----------------------------------------------------------------------------------------------------------------------
#  compute_expansion_nozzle_performance
# ----------------------------------------------------------------------------------------------------------------------        
def compute_expansion_nozzle_performance(expansion_nozzle,nozzle_conditions,conditions):
    """ This computes the output values from the input values according to
    equations from the source. The following properties are computed: 
    expansion_nozzle.outputs.
      stagnation_temperature  (numpy.ndarray): [K]  
      stagnation_pressure     (numpy.ndarray): [Pa]
      stagnation_enthalpy     (numpy.ndarray): [J/kg]
      mach_number             (numpy.ndarray): [-]
      static_temperature      (numpy.ndarray): [K]
      static_enthalpy         (numpy.ndarray): [J/kg]
      velocity                (numpy.ndarray): [m/s]
      static_pressure         (numpy.ndarray): [Pa]
      area_ratio              (numpy.ndarray): [-]
      denisty                 (numpy.ndarray): [kg/m^3] 

    Assumptions:
        1. Constant polytropic efficiency and pressure ratio
        2. If pressures make the Mach number go negative, these values are corrected

    Source:
        https://web.stanford.edu/~cantwell/AA283_Course_Material/AA283_Course_Notes/

    Args:
        conditions.freestream.
          isentropic_expansion_factor         (numpy.ndarray): [-]
          specific_heat_at_constant_pressure  (numpy.ndarray): [J/(kg K)]
          pressure                            (numpy.ndarray): [Pa]
          stagnation_pressure                 (numpy.ndarray): [Pa]
          stagnation_temperature              (numpy.ndarray): [K]
          specific_gas_constant               (numpy.ndarray): [J/(kg K)] 
          mach_number                         (numpy.ndarray): [-]
        expansion_nozzle
          .inputs.stagnation_temperature       (numpy.ndarray): [K]
          .inputs.stagnation_pressure          (numpy.ndarray): [Pa] 
          .pressure_ratio                      (numpy.ndarray): [-]
          .polytropic_efficiency               (numpy.ndarray): [-]    

    Returns:
        None 

    """                 
    # Unpack flight conditions     
    M0       = conditions.freestream.mach_number
    P0       = conditions.freestream.pressure
    Pt0      = conditions.freestream.stagnation_pressure
    Tt0      = conditions.freestream.stagnation_temperature
    
    # Unpack exansion nozzle inputs
    Tt_in    = nozzle_conditions.inputs.stagnation_temperature
    Pt_in    = nozzle_conditions.inputs.stagnation_pressure 
    PR       = expansion_nozzle.pressure_ratio
    etapold  = expansion_nozzle.polytropic_efficiency
    
    P_in  = nozzle_conditions.inputs.static_pressure    
    T_in  = nozzle_conditions.inputs.static_temperature
    
    # Unpack ram inputs       
    working_fluid  = expansion_nozzle.working_fluid 
    gamma          = working_fluid.compute_gamma(T_in,P_in) 
    Cp             = working_fluid.compute_cp(T_in,P_in)    
     
    # Compute output stagnation quantities
    Pt_out   = Pt_in*PR
    Tt_out   = Tt_in*PR**((gamma-1)/(gamma)*etapold)
    ht_out   = Cp*Tt_out
    
    # A cap so pressure doesn't go negative
    Pt_out[Pt_out<P0] = P0[Pt_out<P0]
    
    # Compute the output Mach number, static quantities and the output velocity
    Mach          = np.sqrt((((Pt_out/P0)**((gamma-1)/gamma))-1)*2/(gamma-1)) 
    
    #initializing the Pout array
    P_out         = np.ones_like(Mach)
    
    # Computing output pressure and Mach number for the case Mach <1.0
    i_low         = Mach < 1.0
    P_out[i_low]  = P0[i_low]
    Mach[i_low]   = np.sqrt((((Pt_out[i_low]/P0[i_low])**((gamma[i_low]-1.)/gamma[i_low]))-1.)*2./(gamma[i_low]-1.))
    
    # Computing output pressure and Mach number for the case Mach >=1.0     
    i_high        = Mach >=1.0   
    Mach[i_high]  = Mach[i_high]/Mach[i_high]
    P_out[i_high] = Pt_out[i_high]/(1.+(gamma[i_high]-1.)/2.*Mach[i_high]*Mach[i_high])**(gamma[i_high]/(gamma[i_high]-1.))
    
    # A cap to make sure Mach doesn't go to zero:
    if np.any(Mach<=0.0):
        warn('Pressures Result in Negative Mach Number, making positive',RuntimeWarning)
        Mach[Mach<=0.0] = 0.001
    
    # Compute the output temperature,enthalpy,velocity and density
    T_out         = Tt_out/(1+(gamma-1)/2*Mach*Mach)
    h_out         = T_out * Cp
    u_out         = np.sqrt(2*(ht_out-h_out))
    #rho_out       = P_out/(R*T_out)
    
    # Compute the freestream to nozzle area ratio  
    area_ratio    = (fm_id(M0,gamma)/fm_id(Mach,gamma)*(1/(Pt_out/Pt0))*(np.sqrt(Tt_out/Tt0)))
    
    #pack computed quantities into outputs
    nozzle_conditions.outputs.area_ratio              = area_ratio
    nozzle_conditions.outputs.mach_number             = Mach
    #nozzle_conditions.outputs.density                 = rho_out
    nozzle_conditions.outputs.velocity                = u_out
    nozzle_conditions.outputs.static_pressure         = P_out
    nozzle_conditions.outputs.static_temperature      = T_out
    nozzle_conditions.outputs.static_enthalpy         = h_out
    nozzle_conditions.outputs.stagnation_temperature  = Tt_out
    nozzle_conditions.outputs.stagnation_pressure     = Pt_out
    nozzle_conditions.outputs.stagnation_enthalpy     = ht_out
    
    return 



# RCAIDE/Library/Methods/Energy/Propulsors/Turboshaft_Propulsor/design_turboshaft.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

# RCAIDE Imports     
import RCAIDE
from RCAIDE.Framework.Core                                           import Data
from RCAIDE.Framework.Mission.Common                                 import Conditions
from RCAIDE.Library.Methods.Propulsors.Converters.Ram                import compute_ram_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Combustor          import compute_combustor_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Compressor         import compute_compressor_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Turbine            import compute_turbine_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Expansion_Nozzle   import compute_expansion_nozzle_performance 
from RCAIDE.Library.Methods.Propulsors.Converters.Compression_Nozzle import compute_compression_nozzle_performance
from RCAIDE.Library.Methods.Propulsors.Turboshaft_Propulsor          import size_core 
from RCAIDE.Library.Methods.Propulsors.Common                        import compute_static_sea_level_performance 

# Python package imports   
import numpy                                                                as np

# ----------------------------------------------------------------------------------------------------------------------  
#  Design Turboshaft
# ----------------------------------------------------------------------------------------------------------------------   
def design_turboshaft(turboshaft):  
    #check if mach number and temperature are passed
    if(turboshaft.design_mach_number==None or turboshaft.design_altitude==None):
        
        #raise an error
        raise NameError('The sizing conditions require an altitude and a Mach number')
    
    else:
        #call the atmospheric model to get the conditions at the specified altitude
        atmosphere = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
        atmo_data  = atmosphere.compute_values(turboshaft.design_altitude,turboshaft.design_isa_deviation)
        planet     = RCAIDE.Library.Attributes.Planets.Earth()
        
        p   = atmo_data.pressure          
        T   = atmo_data.temperature       
        rho = atmo_data.density          
        a   = atmo_data.speed_of_sound    
        mu  = atmo_data.dynamic_viscosity   
    
        # setup conditions
        conditions = RCAIDE.Framework.Mission.Common.Results()
    
        # freestream conditions    
        conditions.freestream.altitude                    = np.atleast_1d(turboshaft.design_altitude)
        conditions.freestream.mach_number                 = np.atleast_1d(turboshaft.design_mach_number)
        conditions.freestream.pressure                    = np.atleast_1d(p)
        conditions.freestream.temperature                 = np.atleast_1d(T)
        conditions.freestream.density                     = np.atleast_1d(rho)
        conditions.freestream.dynamic_viscosity           = np.atleast_1d(mu)
        conditions.freestream.gravity                     = np.atleast_1d(planet.compute_gravity(turboshaft.design_altitude))
        conditions.freestream.isentropic_expansion_factor = np.atleast_1d(turboshaft.working_fluid.compute_gamma(T,p))
        conditions.freestream.Cp                          = np.atleast_1d(turboshaft.working_fluid.compute_cp(T,p))
        conditions.freestream.R                           = np.atleast_1d(turboshaft.working_fluid.gas_specific_constant)
        conditions.freestream.speed_of_sound              = np.atleast_1d(a)
        conditions.freestream.velocity                    = np.atleast_1d(a*turboshaft.design_mach_number)
         
         
    fuel_line                = RCAIDE.Library.Components.Energy.Distributors.Fuel_Line()
    segment                  = RCAIDE.Framework.Mission.Segments.Segment()  
    segment.state.conditions = conditions
    segment.state.conditions.energy[fuel_line.tag] = Conditions()
    segment.state.conditions.noise[fuel_line.tag] = Conditions()
    turboshaft.append_operating_conditions(segment) 
    for tag, item in  turboshaft.items(): 
        if issubclass(type(item), RCAIDE.Library.Components.Component):
            item.append_operating_conditions(segment,turboshaft) 
         
    ram                                                   = turboshaft.ram
    inlet_nozzle                                          = turboshaft.inlet_nozzle
    compressor                                            = turboshaft.compressor
    combustor                                             = turboshaft.combustor
    high_pressure_turbine                                 = turboshaft.high_pressure_turbine
    low_pressure_turbine                                  = turboshaft.low_pressure_turbine
    core_nozzle                                           = turboshaft.core_nozzle  

    # unpack component conditions
    turboshaft_conditions   = conditions.energy[turboshaft.tag]
    ram_conditions          = turboshaft_conditions[ram.tag]     
    inlet_nozzle_conditions = turboshaft_conditions[inlet_nozzle.tag]
    core_nozzle_conditions  = turboshaft_conditions[core_nozzle.tag] 
    compressor_conditions   = turboshaft_conditions[compressor.tag]  
    combustor_conditions    = turboshaft_conditions[combustor.tag]
    lpt_conditions          = turboshaft_conditions[low_pressure_turbine.tag]
    hpt_conditions          = turboshaft_conditions[high_pressure_turbine.tag] 
     
    # Step 1: Set the working fluid to determine the fluid properties
    ram.working_fluid                             = turboshaft.working_fluid

    # Step 2: Compute flow through the ram , this computes the necessary flow quantities and stores it into conditions
    compute_ram_performance(ram,ram_conditions,conditions)

    # Step 3: link inlet nozzle to ram 
    inlet_nozzle_conditions.inputs.stagnation_temperature             = ram_conditions.outputs.stagnation_temperature
    inlet_nozzle_conditions.inputs.stagnation_pressure                = ram_conditions.outputs.stagnation_pressure
    inlet_nozzle_conditions.inputs.static_temperature                 = ram_conditions.outputs.static_temperature
    inlet_nozzle_conditions.inputs.static_pressure                    = ram_conditions.outputs.static_pressure
    inlet_nozzle_conditions.inputs.mach_number                        = ram_conditions.outputs.mach_number
    inlet_nozzle.working_fluid                                        = ram.working_fluid

    # Step 4: Compute flow through the inlet nozzle
    compute_compression_nozzle_performance(inlet_nozzle,inlet_nozzle_conditions,conditions)      

    # Step 5: Link low pressure compressor to the inlet nozzle 
    compressor_conditions.inputs.stagnation_temperature  = inlet_nozzle_conditions.outputs.stagnation_temperature
    compressor_conditions.inputs.stagnation_pressure     = inlet_nozzle_conditions.outputs.stagnation_pressure
    compressor_conditions.inputs.static_temperature      = inlet_nozzle_conditions.outputs.static_temperature
    compressor_conditions.inputs.static_pressure         = inlet_nozzle_conditions.outputs.static_pressure
    compressor_conditions.inputs.mach_number             = inlet_nozzle_conditions.outputs.mach_number  
    compressor.working_fluid                             = inlet_nozzle.working_fluid 

    # Step 6: Compute flow through the low pressure compressor
    compute_compressor_performance(compressor,compressor_conditions,conditions)
    
    # Step 11: Link the combustor to the high pressure compressor
    combustor_conditions.inputs.stagnation_temperature                = compressor_conditions.outputs.stagnation_temperature
    combustor_conditions.inputs.stagnation_pressure                   = compressor_conditions.outputs.stagnation_pressure
    combustor_conditions.inputs.static_temperature                    = compressor_conditions.outputs.static_temperature
    combustor_conditions.inputs.static_pressure                       = compressor_conditions.outputs.static_pressure
    combustor_conditions.inputs.mach_number                           = compressor_conditions.outputs.mach_number  
    combustor.working_fluid                                           = compressor.working_fluid 
    
    # Step 12: Compute flow through the high pressor compressor 
    compute_combustor_performance(combustor,combustor_conditions,conditions)

    #link the high pressure turbione to the combustor 
    hpt_conditions.inputs.stagnation_temperature    = combustor_conditions.outputs.stagnation_temperature
    hpt_conditions.inputs.stagnation_pressure       = combustor_conditions.outputs.stagnation_pressure
    hpt_conditions.inputs.fuel_to_air_ratio         = combustor_conditions.outputs.fuel_to_air_ratio 
    hpt_conditions.inputs.static_temperature        = combustor_conditions.outputs.static_temperature
    hpt_conditions.inputs.static_pressure           = combustor_conditions.outputs.static_pressure
    hpt_conditions.inputs.mach_number               = combustor_conditions.outputs.mach_number 
    hpt_conditions.inputs.compressor                = compressor_conditions.outputs 
    hpt_conditions.inputs.bypass_ratio              = 0.0
    hpt_conditions.inputs.fan                       = Data()
    hpt_conditions.inputs.fan.work_done             = 0.0
    high_pressure_turbine.working_fluid             = combustor.working_fluid
    
    compute_turbine_performance(high_pressure_turbine,hpt_conditions,conditions)
            
    #link the low pressure turbine to the high pressure turbine 
    lpt_conditions.inputs.stagnation_temperature     = hpt_conditions.outputs.stagnation_temperature
    lpt_conditions.inputs.stagnation_pressure        = hpt_conditions.outputs.stagnation_pressure 
    lpt_conditions.inputs.static_temperature         = hpt_conditions.outputs.static_temperature
    lpt_conditions.inputs.static_pressure            = hpt_conditions.outputs.static_pressure 
    lpt_conditions.inputs.mach_number                = hpt_conditions.outputs.mach_number  
    low_pressure_turbine.working_fluid               = high_pressure_turbine.working_fluid    
    lpt_conditions.inputs.compressor                 = Data()
    lpt_conditions.inputs.compressor.work_done       = 0.0     
    lpt_conditions.inputs.fuel_to_air_ratio          = combustor_conditions.outputs.fuel_to_air_ratio 
    lpt_conditions.inputs.bypass_ratio               = 0.0
    lpt_conditions.inputs.fan                        = Data()
    lpt_conditions.inputs.fan.work_done              = 0.0
     
    compute_turbine_performance(low_pressure_turbine,lpt_conditions,conditions)
    
    #link the core nozzle to the low pressure turbine
    core_nozzle_conditions.inputs.stagnation_temperature     = lpt_conditions.outputs.stagnation_temperature
    core_nozzle_conditions.inputs.stagnation_pressure        = lpt_conditions.outputs.stagnation_pressure
    core_nozzle_conditions.inputs.static_temperature         = lpt_conditions.outputs.static_temperature
    core_nozzle_conditions.inputs.static_pressure            = lpt_conditions.outputs.static_pressure  
    core_nozzle_conditions.inputs.mach_number                = lpt_conditions.outputs.mach_number   
    core_nozzle.working_fluid                                = low_pressure_turbine.working_fluid  
    
    #flow through the core nozzle
    compute_expansion_nozzle_performance(core_nozzle,core_nozzle_conditions,conditions)

    # compute the thrust using the thrust component
    #link the thrust component to the core nozzle
    turboshaft_conditions.core_exit_velocity                  = core_nozzle_conditions.outputs.velocity
    turboshaft_conditions.core_area_ratio                     = core_nozzle_conditions.outputs.area_ratio
    turboshaft_conditions.core_nozzle                         = core_nozzle_conditions.outputs
    
    #link the thrust component to the combustor
    turboshaft_conditions.fuel_to_air_ratio                   = combustor_conditions.outputs.fuel_to_air_ratio
    
    #link the thrust component to the low pressure compressor
    turboshaft_conditions.combustor_stagnation_temperature    = combustor_conditions.outputs.stagnation_temperature
    turboshaft_conditions.stag_temp_lpt_exit                  = compressor_conditions.inputs.stagnation_temperature
    turboshaft_conditions.stag_press_lpt_exit                 = compressor_conditions.inputs.stagnation_pressure 
    turboshaft_conditions.total_temperature_reference         = compressor_conditions.inputs.stagnation_temperature
    turboshaft_conditions.total_pressure_reference            = compressor_conditions.inputs.stagnation_pressure  

    #compute the power
    turboshaft_conditions.fan_nozzle                          = Data()
    turboshaft_conditions.fan_nozzle.velocity                 = 0.0
    turboshaft_conditions.fan_nozzle.area_ratio               = 0.0
    turboshaft_conditions.fan_nozzle.static_pressure          = 0.0
    turboshaft_conditions.bypass_ratio                        = 0.0
    turboshaft_conditions.flow_through_core                   = 1.0 #scaled constant to turn on core power computation
    turboshaft_conditions.flow_through_fan                    = 0.0 #scaled constant to turn on fan power computation      
    
    # Step 25: Size the core of the turboshaft  
    size_core(turboshaft,turboshaft_conditions,conditions)
    
    # Step 26: Static Sea Level Thrust 
    compute_static_sea_level_performance(turboshaft)
     
    return      
  


# RCAIDE/Library/Components/Energy/Networks/Distribution/Coolant_Line.py 
# 
# Created:  Aug 2024, S. Shekar

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports  
from RCAIDE.Library.Components                                import Component
from RCAIDE.Library.Components.Component                      import Container    

# ----------------------------------------------------------------------------------------------------------------------
#  Coolant Line
# ---------------------------------------------------------------------------------------------------------------------- 
class Coolant_Line(Component):
    """
    Class for modeling coolant distribution lines in thermal management systems
    
    This class represents coolant distribution lines that connect various thermal 
    management components like heat exchangers, reservoirs, and battery cooling systems.

    Attributes
    ----------
    tag : str
        Identifier for the coolant line (default: 'coolant_line')
        
    heat_exchangers : Container
        Collection of heat exchangers connected to the coolant line
        
    reservoirs : Container
        Collection of coolant reservoirs in the system
        
    active : bool
        Flag indicating if the coolant line is operational (default: True)
        
    efficiency : float
        Distribution efficiency of the coolant line (default: 1.0)
        
    battery_modules : Container, optional
        Collection of battery cooling systems, created when batteries are present
        
    identical_battery_modules : bool, optional
        Flag indicating if all battery modules use identical cooling systems

    Notes
    -----
    The coolant line serves as a connection framework between thermal management
    components. It manages the routing of coolant between heat sources (like batteries)
    and heat sinks (like heat exchangers and reservoirs).

    **Definitions**

    'Container'
        A specialized dictionary-like object for storing RCAIDE components
        
    'Heat Acquisition System'
        Components and geometry for removing heat from battery modules

    See Also
    --------
    RCAIDE.Library.Components.Energy.Distributors.Heat_Exchanger
        Heat exchanger components for thermal management
    RCAIDE.Library.Components.Energy.Distributors.Reservoir
        Coolant reservoir components
    """
    
    def __defaults__(self):
        """This sets the default values.
    
        Assumptions:
            None
        
        Source:
            None
        """          
        self.tag                            = 'coolant_line' 
        self.heat_exchangers                = Container()
        self.reservoirs                     = Container() 

                    
    def __init__ (self, distributor=None):
        """
        Initialize coolant line and set up containers for thermal management components
        
        This method initializes empty containers for heat acquisition systems for battery
        modules present on a particular bus.

        Parameters
        ----------
        distributor : Component, optional
            Component containing battery modules that need thermal management
            
        Notes
        -----
        When a distributor with battery modules is provided, the method creates
        containers to store the cooling system components for each battery.
        """               
        self.active                        = True 
        self.efficiency                    = 1.0
        if distributor is not None:
            for tag, item in  distributor.items():
                self.identical_battery_modules  =  distributor.identical_battery_modules
                if tag == 'battery_modules':
                    if not hasattr(self, 'battery_modules'):
                        self.battery_modules = Container()
                    for battery in item:
                        self.battery_modules[battery.tag] = Container()


# RCAIDE/Library/Attributes/Solids/Aluminum.py
# 
 

# Created: Mar 2024 M. Clarke

#-------------------------------------------------------------------------------
# Imports
#-------------------------------------------------------------------------------
from RCAIDE.Framework.Core import Units
from .Solid import Solid 

#-------------------------------------------------------------------------------
# Aluminum for WavyChannel for battery cooling
#------------------------------------------------------------------------------- 
class Aluminum(Solid): 
    """ Physical Constants Specific to 6061-T6 Aluminum
    """

    def __defaults__(self):
        """Sets material properties at instantiation. 

        Assumptions:
            None
    
        Source:
            Cao W, Zhao C, Wang Y, et al. Thermal modeling of full-size-scale cylindrical battery pack cooled
            by channeled liquid flow[J]. International journal of heat and mass transfer, 2019, 138: 1178-1187. 
        """

        self.density                    = 2719
        self.thermal_conductivity       = 202.4
        self.specific_heat_capacity     = 871 
        self.ultimate_tensile_strength  = 310e6 * Units.Pa
        self.ultimate_shear_strength    = 206e6 * Units.Pa
        self.ultimate_bearing_strength  = 607e6 * Units.Pa
        self.yield_tensile_strength     = 276e6 * Units.Pa
        self.yield_shear_strength       = 206e6 * Units.Pa
        self.yield_bearing_strength     = 386e6 * Units.Pa
        self.minimum_gage_thickness     = 0.0   * Units.m
        self.density                    = 2700. * Units['kg/(m**3)']        



# RCAIDE/Methods/Noise/Frequency_Domain_Buildup/Rotor/noise_directivities.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

# Python Package imports  
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
# Compute Noise Directivities  
# ---------------------------------------------------------------------------------------------------------------------- 
def noise_directivities(Theta_er,Phi_er,cos_zeta_r,M_tot):
    '''This computes the laminar boundary layer compoment of broadband noise using the method outlined by the 
    Brooks, Pope and Marcolini (BPM) Model
    
    Assumptions:
        BPM models assumes a naca 0012 airfol  
        
    Source:  
        BPM Model:  Brooks, Thomas F., D. Stuart Pope, and Michael A.
        Marcolini. Airfoil self-noise and prediction. No. L-16528. 1989.
    
    Inputs:  
       Theta_e      - Radiation angle with respect to free stream x (chordwise) [rad]
       Phi_e        - Radiation angle with respect to free stream y (spanwise)  [rad]
       M_c          - Convection Mach number                                    [-]  
       M            - Mach number                                               [-] 
    
    Outputs 
       Dbar_h       - high frequency directivity term                           [-]
       Dbar_l       - low frequency directivity term                            [-] 
       
    Properties Used:
       N/A   
    '''      
    Dbar_h   = (2*(np.sin(Theta_er/2)**2)*((np.sin(Phi_er))**2) )/((1 - M_tot*cos_zeta_r)**4)  # eqn 20 Brooks & Burley
    Dbar_l   = ((np.sin(Theta_er)**2)*((np.sin(Phi_er))**2) )/((1 - M_tot*cos_zeta_r)**4)  # eqn 19 Brooks & Burley  
    return Dbar_h,Dbar_l 


# RCAIDE/Methods/Energy/Propulsors/Converters/DC_Motor/dc_motor_performance.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------    
# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  dc_motor_performance
# ----------------------------------------------------------------------------------------------------------------------           
def compute_Q_from_omega_and_V(motor,motor_conditions,freestream):
    """Calculates the motor's torque

    Assumptions:

    Source:
    N/A

    Inputs:

    Outputs:
    motor_conditions.outputs.torque    [N-m] 

    Properties Used:
    motor.
      gear_ratio           [-]
      speed_constant       [radian/s/V]
      resistance           [ohm]
      outputs.omega        [radian/s]
      gearbox_efficiency   [-]
      expected_current     [A]
      no_load_current      [A]
      inputs.volage        [V]
    """
    
    Res   = motor.resistance
    etaG  = motor.gearbox_efficiency
    exp_i = motor.expected_current
    io    = motor.no_load_current + exp_i*(1-etaG)
    G     = motor.gear_ratio
    Kv    = motor.speed_constant/G
    v     = motor_conditions.inputs.voltage
    omega = motor_conditions.inputs.omega
    
    # Torque
    Q = ((v-omega/Kv)/Res -io)/Kv
    
    motor_conditions.outputs.torque = Q
    motor_conditions.outputs.omega  = omega

    return


# ----------------------------------------------------------------------------------------------------------------------
#  compute_omega_and_Q_from_Cp_and_V
# ----------------------------------------------------------------------------------------------------------------------           
def compute_omega_and_Q_from_Cp_and_V(motor,motor_conditions,freestream):
    """Calculates the motor's rotation rate

    Assumptions:
    Cp (power coefficient) is constant

    Source:
    N/A

    Inputs:
    conditions.
      freestream.velocity                    [m/s]
      freestream.density                     [kg/m^3]
      propulsion.propeller_power_coefficient [-]
    motor_conditions.inputs.voltage                      [V]

    Outputs:
    motor_conditions.outputs.
      torque                                 [Nm]
      omega                                  [radian/s]

    Properties Used:
    motor.
      resistance                             [ohms]
      gearbox_efficiency                     [-]
      expected_current                       [A]
      no_load_current                        [A]
      gear_ratio                             [-]
      speed_constant                         [radian/s/V]
      propeller_radius                       [m]
    """           
    # Unpack 
    rho   = freestream.density[:,0,None]
    Res   = motor.resistance
    etaG  = motor.gearbox_efficiency
    exp_i = motor.expected_current
    io    = motor.no_load_current + exp_i*(1-etaG)
    G     = motor.gear_ratio
    Kv    = motor.speed_constant/G
    R     = motor.rotor_radius
    v     = motor_conditions.inputs.voltage
    Cp    = motor_conditions.inputs.rotor_CP
    

    # Omega
    # This is solved by setting the torque of the motor equal to the torque of the prop
    # It assumes that the Cp is constant
    omega1  =   ((np.pi**(3./2.))*((- 16.*Cp*io*rho*(Kv*Kv*Kv)*(R*R*R*R*R)*(Res*Res) +
                16.*Cp*rho*v*(Kv*Kv*Kv)*(R*R*R*R*R)*Res + (np.pi*np.pi*np.pi))**(0.5) - 
                np.pi**(3./2.)))/(8.*Cp*(Kv*Kv)*(R*R*R*R*R)*Res*rho)
    omega1[np.isnan(omega1)] = 0.0
    
    Q = ((v-omega1/Kv)/Res -io)/Kv
    # store to outputs 
    
    motor_conditions.outputs.torque = Q
    motor_conditions.outputs.omega = omega1

    return


# ----------------------------------------------------------------------------------------------------------------------
# compute_I_from_omega_and_V
# ----------------------------------------------------------------------------------------------------------------------           
def compute_I_from_omega_and_V(motor,motor_conditions,freestream):
    """Calculates the motor's current

    Assumptions:

    Source:
    N/A

    Inputs:
    motor_conditions.inputs.voltage   [V]

    Outputs:
    motor_conditions.outputs.current  [A]
    conditions.
      propulsion.etam      [-] 

    Properties Used:
    motor.
      gear_ratio           [-]
      speed_constant       [radian/s/V]
      resistance           [ohm]
      outputs.omega        [radian/s]
      gearbox_efficiency   [-]
      expected_current     [A]
      no_load_current      [A]
    """                      
    
    # Unpack
    G     = motor.gear_ratio
    Kv    = motor.speed_constant
    Res   = motor.resistance
    v     = motor_conditions.inputs.voltage
    omeg  = motor_conditions.outputs.omega*G
    etaG  = motor.gearbox_efficiency
    exp_i = motor.expected_current
    io    = motor.no_load_current + exp_i*(1-etaG)
    
    i=(v-omeg/Kv)/Res
    
    # This line means the motor cannot recharge the battery
    i[i < 0.0] = 0.0

    # Pack
    motor_conditions.outputs.current    = i
    motor_conditions.outputs.efficiency = (1-io/i)*(1-i*Res/v)
    return



# ----------------------------------------------------------------------------------------------------------------------
# compute_V_and_I_from_omega_and_Kv
# ----------------------------------------------------------------------------------------------------------------------            
def compute_V_and_I_from_omega_and_Kv(motor,motor_conditions,freestream):
    """Calculates the motor's voltage and current

    Assumptions:

    Source:
    N/A

    Inputs:

    Outputs:
    motor_conditions.outputs.current   [A]
    conditions.
      propulsion.volage    [V]
    conditions.
      propulsion.etam      [-] 

    Properties Used:
    motor.
      gear_ratio           [-]
      speed_constant       [radian/s/V]
      resistance           [ohm]
      outputs.omega        [radian/s]
      gearbox_efficiency   [-]
      expected_current     [A]
      no_load_current      [A]
    """                      
           
    Res   = motor.resistance
    etaG  = motor.gearbox_efficiency
    exp_i = motor.expected_current
    io    = motor.no_load_current + exp_i*(1-etaG)
    G     = motor.gear_ratio
    kv    = motor.speed_constant/G
    Q     = motor_conditions.inputs.torque
    omega = motor_conditions.inputs.omega        
    
    v = (Q*kv+io)*Res + omega/kv
    i = (v-omega/kv)/Res
    
    motor_conditions.outputs.voltage    = v
    motor_conditions.outputs.current    = i
    motor_conditions.outputs.efficiency = (1-io/i)*(1-i*Res/v)
    
    return


# RCAIDE/Library/Methods/Weights/Buildups/Common/compute_fuselage_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
# RCAIDE imports   
from RCAIDE.Library.Attributes.Materials import Bidirectional_Carbon_Fiber, Carbon_Fiber_Honeycomb, Paint, Unidirectional_Carbon_Fiber, Acrylic, Steel

# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
# Compute boom weight
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_fuselage_weight(fuse,maxSpan, MTOW, 
             maximum_g_load = 3.8,
             landing_impact_factor = 3.5,
             safety_factor = 1.5):
    """ Calculates the structural mass of a fuselage for an eVTOL vehicle,
        assuming a structural keel taking bending an torsional loads.
        
        Assumptions:  
        If vehicle model does not have material properties assigned, appropriate
        assumptions are made based on RCAIDE's Solids Attributes library.
        
        Sources:
        Project Vahana Conceptual Trade Study

        Inputs:

            config                      RCAIDE Vehicle Configuration
            max_g_load                  Max Accelerative Load During Flight [Unitless]
            landing_impact_factor       Maximum Load Multiplier on Landing  [Unitless]

        Outputs:

            weight:                 Estimated Fuselage Mass             [kg]
        
        Properties Used:
        Material Properties of Imported RCAIDE Solids
    """

    #-------------------------------------------------------------------------------
    # Unpack Inputs
    #------------------------------------------------------------------------------- 
    fLength = fuse.lengths.total
    fWidth  = fuse.width
    fHeight = fuse.heights.maximum
    G_max   = maximum_g_load
    LIF     = landing_impact_factor
    SF      = safety_factor

    #-------------------------------------------------------------------------------
    # Unpack Material Properties
    #-------------------------------------------------------------------------------

    try:
        rbmMat = fuse.keel_materials.root_bending_moment_carrier
    except AttributeError:
        rbmMat = Unidirectional_Carbon_Fiber()
    rbmDen = rbmMat.density
    rbmUTS = rbmMat.ultimate_tensile_strength

    try:
        shearMat = fuse.keel_materials.shear_carrier
    except AttributeError:
        shearMat = Bidirectional_Carbon_Fiber()
    shearDen = shearMat.density
    shearUSS = shearMat.ultimate_shear_strength

    try:
        bearingMat = fuse.keel_materials.bearing_carrier
    except AttributeError:
        bearingMat = Bidirectional_Carbon_Fiber()
    bearingDen = bearingMat.density
    bearingUBS = bearingMat.ultimate_bearing_strength

    try:
        boltMat = fuse.materials.bolt_materials.landing_pad_bolt
    except AttributeError:
        boltMat = Steel()
    boltUSS = boltMat.ultimate_shear_strength


    # Calculate Skin & Canopy Weight Per Unit Area (arealWeight) based on material

    try:
        skinArealWeight = np.sum([(mat.minimum_gage_thickness * mat.density) for mat in fuse.skin_materials])
    except AttributeError:
        skinArealWeight = 1.2995 # Stack of bidirectional CFRP, Honeycomb Core, Paint

    try:
        canopyArealWeight = np.sum([(mat.minimum_gage_thickness * mat.density) for mat in fuse.canopy_materials])
    except AttributeError:
        canopyArealWeight = 3.7465 # Acrylic

    # Calculate fuselage area (using assumption of ellipsoid), and weight:

    S_wet = 4 * np.pi * (((fLength * fWidth/4)**1.6
        + (fLength * fHeight/4)**1.6
        + (fWidth * fHeight/4)**1.6)/3)**(1/1.6)
    skinMass = S_wet * skinArealWeight

    # Calculate the mass of a structural bulkhead

    bulkheadMass = 3 * np.pi * fHeight * fWidth/4 * skinArealWeight

    # Calculate the mass of a canopy

    canopyMass = S_wet/8 * canopyArealWeight

    # Calculate keel mass needed to carry lifting moment

    L_max       = G_max * MTOW * 9.8 * SF  # Max Lifting Load
    M_lift      = L_max * fLength/2.       # Max Moment Due to Lift
    beamWidth   = fWidth/3.                # Allowable Keel Width
    beamHeight  = fHeight/10.              # Allowable Keel Height

    beamArea    = M_lift * beamHeight/(4*rbmUTS*(beamHeight/2)**2)
    massKeel    = beamArea * fLength * rbmDen

    # Calculate keel mass needed to carry wing bending moment shear

    M_bend      = L_max/2 * maxSpan/2                           # Max Bending Moment
    beamArea    = beamHeight * beamWidth                        # Enclosed Beam Area
    beamThk     = 0.5 * M_bend/(shearUSS * beamArea)            # Beam Thickness
    massKeel   += 2*(beamHeight + beamWidth)*beamThk*shearDen

    # Calculate keel mass needed to carry landing impact load assuming

    F_landing   = SF * MTOW * 9.8 * LIF * 0.6403        # Side Landing Force
    boltArea    = F_landing/boltUSS                     # Required Bolt Area
    boltDiam    = 2 * np.sqrt(boltArea/np.pi)           # Bolt Diameter
    lamThk      = F_landing/(boltDiam*bearingUBS)       # Laminate Thickness
    lamVol      = (np.pi*(20*lamThk)**2)*(lamThk/3)     # Laminate Pad volume
    massKeel   += 4*lamVol*bearingDen                   # Mass of 4 Pads

    # Calculate total mass as the sum of skin mass, bulkhead mass, canopy pass,
    # and keel mass. Called weight by RCAIDE convention

    weight = skinMass + bulkheadMass + canopyMass + massKeel

    return weight


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/BWB/compute_cabin_weight.py
# 
# Created: Sep 2024, M. Clarke 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ---------------------------------------------------------------------------------------------------------------------- 

from RCAIDE.Framework.Core import Units

# ---------------------------------------------------------------------------------------------------------------------- 
#  Cabin Weight 
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_cabin_weight(cabin_area, TOGW):
    """ Weight estimate for the cabin (forward section of centerbody) of a BWB.
    Regression from FEA by K. Bradley (George Washington University).
    
    Assumptions:
        -The centerbody uses a pressurized sandwich composite structure
        -Ultimate cabin pressure differential of 18.6psi
        -Critical flight condition: +2.5g maneuver at maximum TOGW
    
    Sources:
        Bradley, K. R., "A Sizing Methodology for the Conceptual Design of 
        Blended-Wing-Body Transports," NASA/CR-2004-213016, 2004.
    
    Inputs:
        cabin_area - the planform area representing the passenger cabin  [meters**2]
        TOGW - Takeoff gross weight of the aircraft                      [kilograms]
    Outputs:
        W_cabin - the estimated structural weight of the BWB cabin      [kilograms]
            
    Properties Used:
    N/A
    """       
    
    # convert to imperial units
    S_cab    = cabin_area / Units.feet ** 2.0
    W        = TOGW       / Units.pounds
    
    W_cabin = 5.698865 * 0.316422 * (W ** 0.166552) * S_cab ** 1.061158
    
    # convert to SI units
    W_cabin = W_cabin * Units.pounds
    
    return W_cabin


# RCAIDE/Library/Methods/Propulsors/Converters/Ducted_Fan/append_ducted_fan_conditions.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_motor_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_ducted_fan_conditions(ducted_fan,segment,energy_conditions): 
    ones_row    = segment.state.ones_row 
    energy_conditions[ducted_fan.tag]                               = Conditions()   
    energy_conditions[ducted_fan.tag].orientation                   = 0. * ones_row(3) 
    energy_conditions[ducted_fan.tag].commanded_thrust_vector_angle = 0. * ones_row(1) 
    energy_conditions[ducted_fan.tag].torque                        = 0. * ones_row(1)
    energy_conditions[ducted_fan.tag].throttle                      = ones_row(1)
    energy_conditions[ducted_fan.tag].thrust                        = 0. * ones_row(1)
    energy_conditions[ducted_fan.tag].rpm                           = 0. * ones_row(1)
    energy_conditions[ducted_fan.tag].omega                         = 0. * ones_row(1)
    energy_conditions[ducted_fan.tag].disc_loading                  = 0. * ones_row(1)                 
    energy_conditions[ducted_fan.tag].power_loading                 = 0. * ones_row(1)
    energy_conditions[ducted_fan.tag].tip_mach                      = 0. * ones_row(1)
    energy_conditions[ducted_fan.tag].efficiency                    = 0. * ones_row(1)
    energy_conditions[ducted_fan.tag].figure_of_merit               = 0. * ones_row(1)
    energy_conditions[ducted_fan.tag].power_coefficient             = 0. * ones_row(1)  
    return 



# RCAIDE/Framework/Analyses/Mission/Segments/Untrimmed/Untrimmed.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports  
from RCAIDE.Framework.Core import Units
from RCAIDE.Framework.Mission.Segments         import Segment
from RCAIDE.Framework.Mission.Common.Results   import Results
from RCAIDE.Library.Mission                    import Common ,Solver,  Segments
from RCAIDE.Framework.Analyses                 import Process   
from RCAIDE.Library.Methods.skip               import skip 

# ----------------------------------------------------------------------------------------------------------------------
#  Untrimmed
# ---------------------------------------------------------------------------------------------------------------------- 

class Untrimmed(Segment):
    """ Base process class used to analyze a vehicle in each flight segment  
    
    Assumptions:
    None

    Source:
    N/A

    Inputs:
    None

    Outputs:
    None

    Properties Used:
    None
    """     
    
    def __defaults__(self):
        """This sets the default values.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """           
        
        # --------------------------------------------------------------
        #   State
        # --------------------------------------------------------------
        
        # conditions
        self.temperature_deviation                   = 0.0
        self.sideslip_angle                          = 0.0 
        self.angle_of_attack                         = 1.0 *  Units.degree
        self.bank_angle                              = 0.0 
        self.linear_acceleration_x                   = 0.
        self.linear_acceleration_y                   = 0.  
        self.linear_acceleration_z                   = 0. # note that down is positive
        self.roll_rate                               = 0.
        self.pitch_rate                              = 0.  
        self.yaw_rate                                = 0.  
        self.state.numerics.number_of_control_points = 2     
        self.trim_lift_coefficient                   = None
        self.state.conditions.update(Results())
        
        # ---------------------------------------------------------------
        # Define Flight Controls and Residuals 
        # ---------------------------------------------------------------     
        self.flight_dynamics_and_controls()    
        
        # --------------------------------------------------------------
        #   Initialize - before iteration
        # -------------------------------------------------------------- 
        initialize                         = self.process.initialize 
        initialize.expand_state            = Solver.expand_state
        initialize.differentials           = Common.Initialize.differentials_dimensionless 
        initialize.conditions              = Segments.Untrimmed.Untrimmed.initialize_conditions  
        
        # --------------------------------------------------------------          
        #   Iterate  
        # -------------------------------------------------------------- 
        iterate                            = self.process.iterate 
        iterate.initials                   = Process()
        iterate.initials.time              = Common.Initialize.time
        iterate.initials.weights           = Common.Initialize.weights
        iterate.initials.energy            = skip
        iterate.initials.inertial_position = Common.Initialize.inertial_position
        iterate.initials.planet_position   = Common.Initialize.planet_position
        
        
        # Unpack Unknowns
        iterate.unknowns                   = Process()
        
        # Update Conditions
        iterate.conditions = Process()
        iterate.conditions.differentials         = Common.Update.differentials_time
        iterate.conditions.orientations          = Common.Update.orientations
        iterate.conditions.acceleration          = skip 
        iterate.conditions.angular_acceleration  = skip 
        iterate.conditions.altitude              = Common.Update.altitude
        iterate.conditions.atmosphere            = Common.Update.atmosphere
        iterate.conditions.gravity               = Common.Update.gravity
        iterate.conditions.freestream            = Common.Update.freestream 
        iterate.conditions.thrust                = Common.Update.thrust
        iterate.conditions.aerodynamics          = Common.Update.aerodynamics
        iterate.conditions.stability             = Common.Update.stability
        iterate.conditions.weights               = Common.Update.weights
        iterate.conditions.forces                = Common.Update.forces
        iterate.conditions.moments               = Common.Update.moments
        iterate.conditions.planet_position       = skip

        # Solve Residuals 
        iterate.unknowns.controls                = Common.Unpack_Unknowns.control_surfaces
        iterate.unknowns.mission                 = Common.Unpack_Unknowns.orientation  
        iterate.residuals.flight_dynamics        = Common.Residuals.flight_dynamics

        # --------------------------------------------------------------  
        #  Post Process   
        # -------------------------------------------------------------- 
        post_process                    = self.process.post_process   
        post_process.inertial_position  = skip
        post_process.energy             = skip
        post_process.noise              = Common.Update.noise
        post_process.emissions          = skip
        
        return 


# RCAIDE/Library/Methods/Aerdoynamics/AERODAS/pre_stall_coefficients.py
# 
# 
# Created:  Jul 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
# Imports 
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Data

# python imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
# Pre Stall Coefficients
# ----------------------------------------------------------------------------------------------------------------------    
def pre_stall_coefficients(state,settings,geometry):
    """Uses the AERODAS method to determine prestall parameters for lift and drag for a single wing

    Assumptions:
    None

    Source:
    NASA TR: "Models of Lift and Drag Coefficients of Stalled and Unstalled Airfoils in
      Wind Turbines and Wind Tunnels" by D. A. Spera

    Inputs:
    state.conditions.aerodynamics.angle_of_attack
    settings.section_zero_lift_angle_of_attack
    geometry.
      section.
        angle_attack_max_prestall_lift
        zero_lift_drag_coefficient
      pre_stall_maximum_drag_coefficient_angle
      pre_stall_maximum_lift_coefficient
      pre_stall_lift_curve_slope 
      pre_stall_maximum_lift_drag_coefficient

    Outputs:
    CL1 (coefficient of lift)                       [Unitless]
    CD1 (coefficient of drag)                       [Unitless]
    (packed in state.conditions.aerodynamics.pre_stall_coefficients[geometry.tag])

    Properties Used:
    N/A
    """  
    
    # unpack inputs
    wing   = geometry
    alpha  = state.conditions.aerodynamics.angle_of_attack * 1.0
    A0     = settings.section_zero_lift_angle_of_attack
    ACL1   = wing.section.angle_attack_max_prestall_lift 
    ACD1   = wing.pre_stall_maximum_drag_coefficient_angle
    CL1max = wing.pre_stall_maximum_lift_coefficient
    S1     = wing.pre_stall_lift_curve_slope  
    CD1max = wing.pre_stall_maximum_lift_drag_coefficient
    CDmin  = wing.section.minimum_drag_coefficient 
    ACDmin = wing.section.minimum_drag_coefficient_angle_of_attack 
    
    if wing.vertical == True:
        alpha = 0. * np.ones_like(alpha)
        
    # Equation 6c
    RCL1          = S1*(ACL1-A0)-CL1max
    RCL1[RCL1<=0] = 1.e-16
    
    # Equation 6d
    N1            = 1 + CL1max/RCL1
    
    # Equation 6a or 6b depending on the alpha
    CL1            = 0.0 * np.ones_like(alpha)
    CL1[alpha>A0]  = S1*(alpha[alpha>A0]-A0)-RCL1[alpha>A0]*((alpha[alpha>A0]-A0)/(ACL1[alpha>A0]-A0))**N1[alpha>A0]
    CL1[alpha==A0] = 0.0
    CL1[alpha<A0]  = S1*(alpha[alpha<A0]-A0)+RCL1[alpha<A0]*((A0-alpha[alpha<A0])/(ACL1[alpha<A0]-A0))**N1[alpha<A0]
    
    # M what is m?
    M              = 2. # Does this need changing

    # Equation 7a
    con      = np.logical_and((2*A0-ACD1)<=alpha,alpha<=ACD1)
    CD1      = np.ones_like(alpha)
    CD1[con] = CDmin[con] + (CD1max[con]-CDmin[con])*((alpha[con] - ACDmin)/(ACD1[con]-ACDmin))**M    
    
    # Equation 7b
    CD1[np.logical_not(con)] = 0.
    
    # Pack outputs
    wing_result = Data(
        lift_coefficient = CL1,
        drag_coefficient = CD1
        )

    state.conditions.aerodynamics.pre_stall_coefficients[wing.tag] =  wing_result
    

    return CL1, CD1


# RCAIDE/Library/Methods/Propulsors/Turbofan_Propulsor/design_turbofan.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jul 2024, RCAIDE Team

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE Imports
import RCAIDE
from RCAIDE.Framework.Mission.Common                                 import Conditions
from RCAIDE.Library.Methods.Propulsors.Converters.Ram                import compute_ram_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Combustor          import compute_combustor_performance 
from RCAIDE.Library.Methods.Propulsors.Converters.Compressor         import compute_compressor_performance 
from RCAIDE.Library.Methods.Propulsors.Converters.Fan                import compute_fan_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Turbine            import compute_turbine_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Expansion_Nozzle   import compute_expansion_nozzle_performance 
from RCAIDE.Library.Methods.Propulsors.Converters.Compression_Nozzle import compute_compression_nozzle_performance
from RCAIDE.Library.Methods.Propulsors.Turbofan_Propulsor            import size_core
from RCAIDE.Library.Methods.Propulsors.Common                        import compute_static_sea_level_performance


# Python package imports
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Design Turbofan
# ---------------------------------------------------------------------------------------------------------------------- 
def design_turbofan(turbofan):
    """Compute perfomance properties of a turbofan based on polytropic ration and combustor properties.
    Turbofan is created by manually linking the different components
    
    
    Assumtions:
       None 
    
    Source:
    
    Args:
        turbofan (dict): turbofan data structure [-]
    
    Returns:
        None 
    
    """
    # check if mach number and temperature are passed
    if(turbofan.design_mach_number==None) and (turbofan.design_altitude==None): 
        raise NameError('The sizing conditions require an altitude and a Mach number') 
    else:
        #call the atmospheric model to get the conditions at the specified altitude
        atmosphere = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
        atmo_data  = atmosphere.compute_values(turbofan.design_altitude,turbofan.design_isa_deviation)
        planet     = RCAIDE.Library.Attributes.Planets.Earth()
        
        p   = atmo_data.pressure          
        T   = atmo_data.temperature       
        rho = atmo_data.density          
        a   = atmo_data.speed_of_sound    
        mu  = atmo_data.dynamic_viscosity           
        U   = a*turbofan.design_mach_number
        # setup conditions
        conditions = RCAIDE.Framework.Mission.Common.Results()
    
        # freestream conditions    
        conditions.freestream.altitude                    = np.atleast_1d(turbofan.design_altitude)
        conditions.freestream.mach_number                 = np.atleast_1d(turbofan.design_mach_number)
        conditions.freestream.pressure                    = np.atleast_1d(p)
        conditions.freestream.temperature                 = np.atleast_1d(T)
        conditions.freestream.density                     = np.atleast_1d(rho)
        conditions.freestream.dynamic_viscosity           = np.atleast_1d(mu)
        conditions.freestream.gravity                     = np.atleast_1d(planet.compute_gravity(turbofan.design_altitude))
        conditions.freestream.isentropic_expansion_factor = np.atleast_1d(turbofan.working_fluid.compute_gamma(T,p))
        conditions.freestream.Cp                          = np.atleast_1d(turbofan.working_fluid.compute_cp(T,p))
        conditions.freestream.R                           = np.atleast_1d(turbofan.working_fluid.gas_specific_constant)
        conditions.freestream.speed_of_sound              = np.atleast_1d(a)
        conditions.freestream.velocity                    = np.atleast_1d(U) 
     
    segment                  = RCAIDE.Framework.Mission.Segments.Segment()  
    segment.state.conditions = conditions 
    turbofan.append_operating_conditions(segment) 
    for tag, item in  turbofan.items(): 
        if issubclass(type(item), RCAIDE.Library.Components.Component):
            item.append_operating_conditions(segment,turbofan) 
    
    ram                       = turbofan.ram
    inlet_nozzle              = turbofan.inlet_nozzle
    fan                       = turbofan.fan
    low_pressure_compressor   = turbofan.low_pressure_compressor
    high_pressure_compressor  = turbofan.high_pressure_compressor
    combustor                 = turbofan.combustor
    high_pressure_turbine     = turbofan.high_pressure_turbine
    low_pressure_turbine      = turbofan.low_pressure_turbine
    core_nozzle               = turbofan.core_nozzle
    fan_nozzle                = turbofan.fan_nozzle 
    bypass_ratio              = turbofan.bypass_ratio  

    # unpack component conditions
    turbofan_conditions     = conditions.energy[turbofan.tag]
    ram_conditions          = turbofan_conditions[ram.tag]    
    inlet_nozzle_conditions = turbofan_conditions[inlet_nozzle.tag]
    fan_conditions          = turbofan_conditions[fan.tag]    
    lpc_conditions          = turbofan_conditions[low_pressure_compressor.tag]
    hpc_conditions          = turbofan_conditions[high_pressure_compressor.tag]
    combustor_conditions    = turbofan_conditions[combustor.tag] 
    lpt_conditions          = turbofan_conditions[low_pressure_turbine.tag]
    hpt_conditions          = turbofan_conditions[high_pressure_turbine.tag]
    core_nozzle_conditions  = turbofan_conditions[core_nozzle.tag]
    fan_nozzle_conditions   = turbofan_conditions[fan_nozzle.tag]    
     
    # Step 1: Set the working fluid to determine the fluid properties
    ram.working_fluid                             = turbofan.working_fluid
    
    # Step 2: Compute flow through the ram , this computes the necessary flow quantities and stores it into conditions
    compute_ram_performance(ram,ram_conditions,conditions)
    
    # Step 3: link inlet nozzle to ram 
    inlet_nozzle_conditions.inputs.stagnation_temperature             = ram_conditions.outputs.stagnation_temperature
    inlet_nozzle_conditions.inputs.stagnation_pressure                = ram_conditions.outputs.stagnation_pressure
    inlet_nozzle_conditions.inputs.static_temperature                 = ram_conditions.outputs.static_temperature
    inlet_nozzle_conditions.inputs.static_pressure                    = ram_conditions.outputs.static_pressure
    inlet_nozzle_conditions.inputs.mach_number                        = ram_conditions.outputs.mach_number
    inlet_nozzle.working_fluid                                        = ram.working_fluid
    
    # Step 4: Compute flow through the inlet nozzle
    compute_compression_nozzle_performance(inlet_nozzle,inlet_nozzle_conditions,conditions)
    
    # Step 5: Link the fan to the inlet nozzle
    fan_conditions.inputs.stagnation_temperature                      = inlet_nozzle_conditions.outputs.stagnation_temperature
    fan_conditions.inputs.stagnation_pressure                         = inlet_nozzle_conditions.outputs.stagnation_pressure
    fan_conditions.inputs.static_temperature                          = inlet_nozzle_conditions.outputs.static_temperature
    fan_conditions.inputs.static_pressure                             = inlet_nozzle_conditions.outputs.static_pressure
    fan_conditions.inputs.mach_number                                 = inlet_nozzle_conditions.outputs.mach_number  
    fan.working_fluid                                                 = inlet_nozzle.working_fluid
     
    # Step 6: Compute flow through the fan
    compute_fan_performance(fan,fan_conditions,conditions)    

    # Step 7: Link low pressure compressor to the inlet nozzle
    lpc_conditions.inputs.stagnation_temperature                      = fan_conditions.outputs.stagnation_temperature
    lpc_conditions.inputs.stagnation_pressure                         = fan_conditions.outputs.stagnation_pressure
    lpc_conditions.inputs.static_temperature                          = fan_conditions.outputs.static_temperature
    lpc_conditions.inputs.static_pressure                             = fan_conditions.outputs.static_pressure
    lpc_conditions.inputs.mach_number                                 = fan_conditions.outputs.mach_number  
    low_pressure_compressor.working_fluid                             = fan.working_fluid
    
    # Step 8: Compute flow through the low pressure compressor
    compute_compressor_performance(low_pressure_compressor,lpc_conditions,conditions)
    
    # Step 9: Link the high pressure compressor to the low pressure compressor
    hpc_conditions.inputs.stagnation_temperature                      = lpc_conditions.outputs.stagnation_temperature
    hpc_conditions.inputs.stagnation_pressure                         = lpc_conditions.outputs.stagnation_pressure
    hpc_conditions.inputs.static_temperature                          = lpc_conditions.outputs.static_temperature
    hpc_conditions.inputs.static_pressure                             = lpc_conditions.outputs.static_pressure
    hpc_conditions.inputs.mach_number                                 = lpc_conditions.outputs.mach_number  
    high_pressure_compressor.working_fluid                            = low_pressure_compressor.working_fluid    
    
    # Step 10: Compute flow through the high pressure compressor
    compute_compressor_performance(high_pressure_compressor,hpc_conditions,conditions)

    # Step 11: Link the combustor to the high pressure compressor    
    combustor_conditions.inputs.stagnation_temperature                = hpc_conditions.outputs.stagnation_temperature
    combustor_conditions.inputs.stagnation_pressure                   = hpc_conditions.outputs.stagnation_pressure
    combustor_conditions.inputs.static_temperature                    = hpc_conditions.outputs.static_temperature
    combustor_conditions.inputs.static_pressure                       = hpc_conditions.outputs.static_pressure
    combustor_conditions.inputs.mach_number                           = hpc_conditions.outputs.mach_number  
    combustor.working_fluid                                           = high_pressure_compressor.working_fluid     
    
    # Step 12: Compute flow through the high pressor compressor 
    compute_combustor_performance(combustor,combustor_conditions,conditions)
    
    # Step 13: Link the high pressure turbione to the combustor
    hpt_conditions.inputs.stagnation_temperature    = combustor_conditions.outputs.stagnation_temperature
    hpt_conditions.inputs.stagnation_pressure       = combustor_conditions.outputs.stagnation_pressure
    hpt_conditions.inputs.fuel_to_air_ratio         = combustor_conditions.outputs.fuel_to_air_ratio 
    hpt_conditions.inputs.static_temperature        = combustor_conditions.outputs.static_temperature
    hpt_conditions.inputs.static_pressure           = combustor_conditions.outputs.static_pressure
    hpt_conditions.inputs.mach_number               = combustor_conditions.outputs.mach_number       
    hpt_conditions.inputs.compressor                = hpc_conditions.outputs  
    hpt_conditions.inputs.fan                       = fan_conditions.outputs 
    hpt_conditions.inputs.bypass_ratio              = 0.0
    high_pressure_turbine.working_fluid             = combustor.working_fluid    
    
    # Step 14: Compute flow through the high pressure turbine
    compute_turbine_performance(high_pressure_turbine,hpt_conditions,conditions)
            
    # Step 15: Link the low pressure turbine to the high pressure turbine
    lpt_conditions.inputs.stagnation_temperature     = hpt_conditions.outputs.stagnation_temperature
    lpt_conditions.inputs.stagnation_pressure        = hpt_conditions.outputs.stagnation_pressure
    lpt_conditions.inputs.static_temperature         = hpt_conditions.outputs.static_temperature
    lpt_conditions.inputs.static_pressure            = hpt_conditions.outputs.static_pressure  
    lpt_conditions.inputs.mach_number                = hpt_conditions.outputs.mach_number    
    low_pressure_turbine.working_fluid               = high_pressure_turbine.working_fluid     
    lpt_conditions.inputs.compressor                 = lpc_conditions.outputs 
    lpt_conditions.inputs.fuel_to_air_ratio          = combustor_conditions.outputs.fuel_to_air_ratio 
    lpt_conditions.inputs.fan                        = fan_conditions.outputs 
    lpt_conditions.inputs.bypass_ratio               = bypass_ratio 
    
    # Step 16: Compute flow through the low pressure turbine
    compute_turbine_performance(low_pressure_turbine,lpt_conditions,conditions)
    
    # Step 17: Link the core nozzle to the low pressure turbine
    core_nozzle_conditions.inputs.stagnation_temperature     = lpt_conditions.outputs.stagnation_temperature
    core_nozzle_conditions.inputs.stagnation_pressure        = lpt_conditions.outputs.stagnation_pressure
    core_nozzle_conditions.inputs.static_temperature         = lpt_conditions.outputs.static_temperature
    core_nozzle_conditions.inputs.static_pressure            = lpt_conditions.outputs.static_pressure  
    core_nozzle_conditions.inputs.mach_number                = lpt_conditions.outputs.mach_number   
    core_nozzle.working_fluid                                = low_pressure_turbine.working_fluid 
    
    # Step 18: Compute flow through the core nozzle
    compute_expansion_nozzle_performance(core_nozzle,core_nozzle_conditions,conditions)
   
    # Step 19: Link the fan nozzle to the fan
    fan_nozzle_conditions.inputs.stagnation_temperature     = fan_conditions.outputs.stagnation_temperature
    fan_nozzle_conditions.inputs.stagnation_pressure        = fan_conditions.outputs.stagnation_pressure
    fan_nozzle_conditions.inputs.static_temperature         = fan_conditions.outputs.static_temperature
    fan_nozzle_conditions.inputs.static_pressure            = fan_conditions.outputs.static_pressure  
    fan_nozzle_conditions.inputs.mach_number                = fan_conditions.outputs.mach_number   
    fan_nozzle.working_fluid                                = fan.working_fluid
    
    # Step 20: Compute flow through the fan nozzle
    compute_expansion_nozzle_performance(fan_nozzle,fan_nozzle_conditions,conditions)
     
    # Step 21: Link the turbofan to outputs from various compoments    
    turbofan_conditions.bypass_ratio                             = bypass_ratio
    turbofan_conditions.fan_nozzle_exit_velocity                 = fan_nozzle_conditions.outputs.velocity
    turbofan_conditions.fan_nozzle_area_ratio                    = fan_nozzle_conditions.outputs.area_ratio  
    turbofan_conditions.fan_nozzle_static_pressure               = fan_nozzle_conditions.outputs.static_pressure
    turbofan_conditions.core_nozzle_area_ratio                   = core_nozzle_conditions.outputs.area_ratio 
    turbofan_conditions.core_nozzle_static_pressure              = core_nozzle_conditions.outputs.static_pressure
    turbofan_conditions.core_nozzle_exit_velocity                = core_nozzle_conditions.outputs.velocity 
    turbofan_conditions.fuel_to_air_ratio                        = combustor_conditions.outputs.fuel_to_air_ratio 
    turbofan_conditions.total_temperature_reference              = lpc_conditions.outputs.stagnation_temperature
    turbofan_conditions.total_pressure_reference                 = lpc_conditions.outputs.stagnation_pressure
    turbofan_conditions.flow_through_core                        = 1./(1.+bypass_ratio) #scaled constant to turn on core thrust computation
    turbofan_conditions.flow_through_fan                         = bypass_ratio/(1.+bypass_ratio) #scaled constant to turn on fan thrust computation        

    # Step 22: Size the core of the turbofan  
    size_core(turbofan,turbofan_conditions,conditions) 
    mass_flow                     = turbofan.mass_flow_rate_design
    turbofan.design_core_massflow = mass_flow   
    
    # Step 23: Static Sea Level Thrust 
    compute_static_sea_level_performance(turbofan)
     
    return 


# RCAIDE/Library/Plots/Geometry/plot_3d_wing.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
import RCAIDE
from RCAIDE.Framework.Core import Data
from RCAIDE.Library.Plots.Geometry.Common.contour_surface_slice import contour_surface_slice
from RCAIDE.Library.Methods.Geometry.Airfoil import import_airfoil_geometry
from RCAIDE.Library.Methods.Geometry.Airfoil import compute_naca_4series 
import numpy as np     

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------  
def plot_3d_wing(plot_data,wing,number_of_airfoil_points = 21, color_map='greys',alpha=1):
    """ This plots the wings of a vehicle

    Assumptions:
    None

    Source:
    None

    Inputs:
    VD.
       XA1...ZB2    - coordinates of wing vortex distribution
    color_map       - color of panel 

    Properties Used:
    N/A
    """
     
    af_pts    = number_of_airfoil_points-1 
    n_segments = len(wing.segments)  
    if n_segments>0:
        dim =  n_segments
    else:
        dim = 2 
 
    G = generate_3d_wing_points(wing,number_of_airfoil_points,dim)
    # ------------------------------------------------------------------------
    # Plot Rotor Blade
    # ------------------------------------------------------------------------
    for sec in range(dim-1):
        for loc in range(af_pts):
            X = np.array([[G.XA1[sec,loc],G.XA2[sec,loc]],
                 [G.XB1[sec,loc],G.XB2[sec,loc]]])
            Y = np.array([[G.YA1[sec,loc],G.YA2[sec,loc]],
                 [G.YB1[sec,loc],G.YB2[sec,loc]]])
            Z = np.array([[G.ZA1[sec,loc],G.ZA2[sec,loc]],
                 [G.ZB1[sec,loc],G.ZB2[sec,loc]]]) 
             
            values      = np.ones_like(X) 
            verts       = contour_surface_slice(X, Y, Z ,values,color_map)
            plot_data.append(verts)
    if wing.symmetric:
        if wing.vertical: 
            for sec in range(dim-1):
                for loc in range(af_pts):
                    X = np.array([[G.XA1[sec,loc],G.XA2[sec,loc]],[G.XB1[sec,loc],G.XB2[sec,loc]]])
                    Y = np.array([[G.YA1[sec,loc],G.YA2[sec,loc]],[G.YB1[sec,loc],G.YB2[sec,loc]]])
                    Z = np.array([[-G.ZA1[sec,loc], -G.ZA2[sec,loc]],[-G.ZB1[sec,loc], -G.ZB2[sec,loc]]]) 
                     
                    values      = np.ones_like(X) 
                    verts       = contour_surface_slice(X, Y, Z ,values,color_map)
                    plot_data.append(verts)
        else:
            for sec in range(dim-1):
                for loc in range(af_pts):
                    X = np.array([[G.XA1[sec,loc],G.XA2[sec,loc]],[G.XB1[sec,loc],G.XB2[sec,loc]]])
                    Y = np.array([[-G.YA1[sec,loc], -G.YA2[sec,loc]], [-G.YB1[sec,loc], -G.YB2[sec,loc]]])
                    Z = np.array([[G.ZA1[sec,loc],G.ZA2[sec,loc]], [G.ZB1[sec,loc],G.ZB2[sec,loc]]]) 
                     
                    values      = np.ones_like(X) 
                    verts       = contour_surface_slice(X, Y, Z ,values,color_map)
                    plot_data.append(verts)
            
             
    return plot_data
 
def generate_3d_wing_points(wing,n_points,dim):
    """ This generates the coordinates of the blade surface for plotting in the aircraft frame (x-back, z-up)

    Assumptions:
    None

    Source:
    None

    Inputs:
    rotor            - RCAIDE rotor
    n_points         - number of points around airfoils of each blade section
    dim              - number for radial dimension
    i                - blade number
    aircraftRefFrame - boolean to convert the coordinates from rotor frame to aircraft frame 

    Properties Used:
    N/A
    """    
    # unpack  
    # obtain the geometry for each segment in a loop                                            
    symm                 = wing.symmetric
    semispan             = wing.spans.projected*0.5 * (2 - symm) 
    root_chord           = wing.chords.root
    segments             = wing.segments
    n_segments           = len(segments.keys()) 
    origin               = wing.origin   
        
    if n_segments > 0: 
        pts              = np.zeros((dim,n_points, 3,1))  
        section_twist    = np.zeros((dim,n_points, 3,3))
        section_twist[:, :, 0, 0] = 1        
        section_twist[:, :, 1, 1] = 1
        section_twist[:, :, 2, 2] = 1 
        translation        = np.zeros((dim,n_points, 3,1)) 
        translation[0, :, 0,:] = origin[0][0]  
        translation[0, :, 1,:] = origin[0][1]  
        translation[0, :, 2,:] = origin[0][2]  
        
        for i in range(n_segments):
            current_seg = list(segments.keys())[i]
            airfoil = wing.segments[current_seg].airfoil 
            if  airfoil !=  None:                 
                if type(airfoil) == RCAIDE.Library.Components.Airfoils.NACA_4_Series_Airfoil:
                    geometry = compute_naca_4series(airfoil.NACA_4_Series_code,n_points)
                elif type(airfoil) == RCAIDE.Library.Components.Airfoils.Airfoil: 
                    geometry     = import_airfoil_geometry(airfoil.coordinate_file,n_points)
            else:
                geometry = compute_naca_4series('0012',n_points)  
            
            if (i == n_segments-1):
                sweep = 0                                 
            else: 
                next_seg = list(segments.keys())[i+1]                
                if wing.segments[current_seg].sweeps.leading_edge is not None: 
                    # If leading edge sweep is defined 
                    sweep       = wing.segments[current_seg].sweeps.leading_edge  
                else:   
                    # If quarter chord sweep is defined, convert it to leading edge sweep
                    sweep_quarter_chord = wing.segments[current_seg].sweeps.quarter_chord 
                    chord_fraction      = 0.25                          
                    segment_root_chord  = root_chord*wing.segments[current_seg].root_chord_percent
                    segment_tip_chord   = root_chord*wing.segments[next_seg].root_chord_percent
                    segment_span        = semispan*(wing.segments[next_seg].percent_span_location - wing.segments[current_seg].percent_span_location )
                    sweep               = np.arctan(((segment_root_chord*chord_fraction) + (np.tan(sweep_quarter_chord )*segment_span - chord_fraction*segment_tip_chord)) /segment_span) 
            dihedral = wing.segments[current_seg].dihedral_outboard    
            twist    = wing.segments[current_seg].twist
            
            if wing.vertical: 
                pts[i,:,0,0]   = geometry.x_coordinates * wing.segments[current_seg].root_chord_percent * wing.chords.root 
                pts[i,:,1,0]   = geometry.y_coordinates * wing.segments[current_seg].root_chord_percent * wing.chords.root 
                pts[i,:,2,0]   = np.zeros_like(geometry.y_coordinates) 
              
                section_twist[i,:,0,0] = np.cos(twist) 
                section_twist[i,:,0,1] = -np.sin(twist)  
                section_twist[i,:,1,0] = np.sin(twist) 
                section_twist[i,:,1,1] = np.cos(twist) 
            
            else: 
                pts[i,:,0,0]   = geometry.x_coordinates * wing.segments[current_seg].root_chord_percent * wing.chords.root
                pts[i,:,1,0]   = np.zeros_like(geometry.y_coordinates) 
                pts[i,:,2,0]   = geometry.y_coordinates * wing.segments[current_seg].root_chord_percent * wing.chords.root  
                                 

                section_twist[i,:,0,0] = np.cos(twist) 
                section_twist[i,:,0,2] = np.sin(twist)  
                section_twist[i,:,2,0] = -np.sin(twist) 
                section_twist[i,:,2,2] =  np.cos(twist)  
             
            if (i != n_segments-1):
                # update origin for next segment 
                segment_percent_span =    wing.segments[next_seg].percent_span_location - wing.segments[current_seg].percent_span_location     
                if wing.vertical:
                    dz = semispan*segment_percent_span
                    dy = dz*np.tan(dihedral)
                    l  = dz/np.cos(dihedral)
                    dx = l*np.tan(sweep)
                else:
                    dy = semispan*segment_percent_span
                    dz = dy*np.tan(dihedral)
                    l  = dy/np.cos(dihedral)
                    dx = l*np.tan(sweep)
                translation[i+1,:,0,:] = translation[i,:,0,:] + dx
                translation[i+1,:,1,:] = translation[i,:,1,:] + dy
                translation[i+1,:,2,:] = translation[i,:,2,:] + dz 
    else:

        pts              = np.zeros((dim,n_points, 3,1))  
        section_twist    = np.zeros((dim,n_points, 3,3))
        section_twist[:, :, 0, 0] = 1        
        section_twist[:, :, 1, 1] = 1
        section_twist[:, :, 2, 2] = 1
        translation      = np.zeros((dim,n_points, 3,1))
        
        airfoil = wing.airfoil 
        if wing.airfoil != None: 
            if type(airfoil) == RCAIDE.Library.Components.Airfoils.NACA_4_Series_Airfoil:
                geometry = compute_naca_4series(airfoil.NACA_4_Series_code,n_points)
            elif type(airfoil) == RCAIDE.Library.Components.Airfoils.Airfoil: 
                geometry     = import_airfoil_geometry(airfoil.coordinate_file,n_points)
        else:
            geometry = compute_naca_4series('0012',n_points)
            
        dihedral              = wing.dihedral
        if wing.sweeps.leading_edge  is not None: 
            sweep      = wing.sweeps.leading_edge
        else:  
            sweep_quarter_chord = wing.sweeps.quarter_chord 
            chord_fraction      = 0.25                          
            segment_root_chord  = wing.chords.root
            segment_tip_chord   = wing.chords.tip
            segment_span        = semispan 
            sweep       = np.arctan(((segment_root_chord*chord_fraction) + (np.tan(sweep_quarter_chord )*segment_span - chord_fraction*segment_tip_chord)) /segment_span)  
           
        # append root section     
        translation[:, :, 0,:] = origin[0][0]  
        translation[:, :, 1,:] = origin[0][1]  
        translation[:, :, 2,:] = origin[0][2] 
       
        if wing.vertical: 
            pts[0,:,0,0]   = geometry.x_coordinates *  wing.chords.root
            pts[0,:,1,0]   = geometry.y_coordinates *  wing.chords.root
            pts[0,:,2,0]   = np.zeros_like(geometry.y_coordinates)
            
            pts[1,:,0,0]   = geometry.x_coordinates *  wing.chords.tip  
            pts[1,:,1,0]   = geometry.y_coordinates *  wing.chords.tip  
            pts[1,:,2,0]   = np.zeros_like(geometry.y_coordinates)   
            
            translation[1, :, 0,:] += semispan*np.tan(sweep)
            translation[1, :, 1,:] += semispan*np.tan(dihedral) 
            translation[1, :, 2,:] += semispan 

            section_twist[0,:,0,0] = np.cos(wing.twists.root) 
            section_twist[0,:,0,1] = -np.sin(wing.twists.root)  
            section_twist[0,:,1,0] = np.sin(wing.twists.root) 
            section_twist[0,:,1,1] = np.cos(wing.twists.root)
             
            section_twist[1,:,0,0] = np.cos(wing.twists.tip) 
            section_twist[1,:,0,1] = -np.sin(wing.twists.tip)  
            section_twist[1,:,1,0] = np.sin(wing.twists.tip) 
            section_twist[1,:,1,1] = np.cos(wing.twists.tip)
            
            
        else:
            pts[0,:,0,0]   = geometry.x_coordinates *  wing.chords.root
            pts[0,:,1,0]   = np.zeros_like(geometry.y_coordinates) 
            pts[0,:,2,0]   = geometry.y_coordinates *  wing.chords.root
            
            pts[1,:,0,0]   = geometry.x_coordinates *  wing.chords.tip  
            pts[1,:,1,0]   = np.zeros_like(geometry.y_coordinates)  
            pts[1,:,2,0]   = geometry.y_coordinates *  wing.chords.tip  
            
    
            translation[1, :, 0,:] +=  semispan*np.tan(sweep)
            translation[1, :, 1,:] += semispan 
            translation[1, :, 2,:] += semispan*np.tan(dihedral)     

            section_twist[0,:,0,0] = np.cos(wing.twists.root) 
            section_twist[0,:,0,2] = np.sin(wing.twists.root)  
            section_twist[0,:,2,0] = -np.sin(wing.twists.root) 
            section_twist[0,:,2,2] =  np.cos(wing.twists.root)
             
            section_twist[1,:,0,0] = np.cos(wing.twists.tip) 
            section_twist[1,:,0,2] = np.sin(wing.twists.tip)  
            section_twist[1,:,2,0] = -np.sin(wing.twists.tip) 
            section_twist[1,:,2,2] =  np.cos(wing.twists.tip) 
 
    mat     = translation + np.matmul(section_twist ,pts)
    
    # ---------------------------------------------------------------------------------------------
    # create empty data structure for storing geometry
    G = Data()

    # store node points
    G.X    = mat[:,:,0,0]  
    G.Y    = mat[:,:,1,0]  
    G.Z    = mat[:,:,2,0]
    G.PTS  = mat[:,:,:,0]

    # store points
    G.XA1  = mat[:-1,:-1,0,0] 
    G.YA1  = mat[:-1,:-1,1,0] 
    G.ZA1  = mat[:-1,:-1,2,0] 
    G.XA2  = mat[:-1,1:,0,0]  
    G.YA2  = mat[:-1,1:,1,0]  
    G.ZA2  = mat[:-1,1:,2,0]  
    G.XB1  = mat[1:,:-1,0,0]  
    G.YB1  = mat[1:,:-1,1,0]  
    G.ZB1  = mat[1:,:-1,2,0]  
    G.XB2  = mat[1:,1:,0,0]   
    G.YB2  = mat[1:,1:,1,0]   
    G.ZB2  = mat[1:,1:,2,0]      
    
    return G
     





## @defgroup Methods-Propulsion-Rotor_Wake-Fidelity_Zero
# compute_wake_contraction_matrix.py
# 
# Created:  Sep 2020, M. Clarke 
#           Jul 2021, E. Botero
#           Sep 2021, R. Erhard
# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

# package imports
import numpy as np 

## @defgroup Methods-Propulsion-Rotor_Wake-Fidelity_Zero
def compute_wake_contraction_matrix(prop,Nr,m,nts,X_pts,prop_outputs):
    """ This computes slipstream development factor for all points 
    along slipstream

    Assumptions: 
    Fixed wake with helical shape  

    Source:  
    Stone, R. Hugh. "Aerodynamic modeling of the wing-propeller 
    interaction for a tail-sitter unmanned air vehicle." Journal 
    of Aircraft 45.1 (2008): 198-210.
    
    Inputs: 
    prop     - propeller/rotor data structure       
    Nr       - discretization on propeller/rotor [Unitless] 
    m        - control points in segemnt         [Unitless] 
    nts      - number of timesteps               [Unitless] 
    X_pts    - location of wake points           [meters] 

    Properties Used:
    N/A
    """    
    r                 = prop.radius_distribution  
    rdim              = Nr-1
    B                 = prop.number_of_blades
    va                = np.mean(prop_outputs.disc_axial_induced_velocity, axis=2)  # induced velocitied averaged around the azimuth
    R0                = prop.hub_radius 
    R_p               = prop.tip_radius  
    s                 = X_pts[0,:,0,-1,:] - prop.origin[0][0]    #  ( control point, blade number,  location on blade, time step)  
    s2                = 1 + s/(np.sqrt(s**2 + R_p**2))
    Kd                = np.repeat(np.atleast_2d(s2)[:, None, :], rdim , axis = 1)  
    
    # TO DO: UPDATE FOR ANGLES SO THAT VELOCITY IS COMPONENT IN ROTOR AXIS FRAME
    VX                = np.repeat(np.repeat(np.atleast_2d(prop_outputs.velocity[:,0]).T, rdim, axis = 1)[:, :, None], nts , axis = 2) # dimension (num control points, propeller distribution, wake points )
   
    prop_dif          = np.atleast_2d(va[:,1:] +  va[:,:-1])
    prop_dif          = np.repeat(prop_dif[:,  :, None], nts, axis=2) 
     
    Kv                = (2*VX + prop_dif) /(2*VX + Kd*prop_dif)  
    
    r_diff            = np.ones((m,rdim))*(r[1:]**2 - r[:-1]**2 )
    r_diff            = np.repeat(np.atleast_2d(r_diff)[:, :, None], nts, axis = 2) 
    r_prime           = np.zeros((m,Nr,nts))                
    r_prime[:,0,:]    = R0   
    for j in range(rdim):
        r_prime[:,1+j,:]   = np.sqrt(r_prime[:,j,:]**2 + (r_diff*Kv)[:,j,:])                               
    
    wake_contraction  = np.repeat((r_prime/np.repeat(np.atleast_2d(r)[:, :, None], nts, axis = 2))[:,None,:,:], B, axis = 1)            
    
    return wake_contraction 
            



# RCAIDE/Library/Missions/Segments/Climb/Constant_Throttle_Constant_Speed.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

# Package imports  
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------
def unpack_body_angle(segment):
    """Unpacks and sets the proper value for body angle

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
    state.unknowns.body_angle                      [Radians]

    Outputs:
    state.conditions.frames.body.inertial_rotation [Radians]

    Properties Used:
    N/A
    """          
    
    ctrls    = segment.assigned_control_variables 

    # Body Angle Control    
    if ctrls.body_angle.active: 
        segment.state.conditions.frames.body.inertial_rotations[:,1] = segment.state.unknowns.body_angle[:,0] 
    else:
        segment.state.conditions.frames.body.inertial_rotations[:,1] = segment.angle_of_attack            

    if ctrls.bank_angle.active: 
        segment.state.conditions.frames.body.inertial_rotations[:,0] = segment.state.unknowns.bank_angle[:,0]
    else:
        segment.state.conditions.frames.body.inertial_rotations[:,0] = segment.bank_angle
            
    segment.state.conditions.frames.body.inertial_rotations[:,2] =  segment.state.conditions.frames.planet.true_heading[:,0]     
         
# ----------------------------------------------------------------------
#  Initialize Conditions
# ----------------------------------------------------------------------

def initialize_conditions(segment):
    """Sets the specified conditions which are given for the segment type.
    
    Assumptions:
    Constant throttle estting, with a constant true airspeed

    Source:
    N/A

    Inputs:
    segment.air_speed                                   [meters/second] 
    segment.altitude_start                              [meters]
    segment.altitude_end                                [meters]
    segment.state.numerics.dimensionless.control_points [Unitless]
    conditions.freestream.density                       [kilograms/meter^3]

    Outputs:
    conditions.frames.inertial.velocity_vector          [meters/second]
    conditions.energy.throttle                          [Unitless]

    Properties Used:
    N/A
    """         
    
    # unpack 
    alt0       = segment.altitude_start 
    v_mag      = segment.air_speed 
    beta       = segment.sideslip_angle
    alpha      = segment.state.unknowns.wind_angle[:,0][:,None]
    theta      = segment.state.unknowns.body_angle[:,0][:,None]
    conditions = segment.state.conditions  

    # check for initial altitude
    if alt0 is None:
        if not segment.state.initials: raise AttributeError('initial altitude not set')
        alt0 = -1.0 *segment.state.initials.conditions.frames.inertial.position_vector[-1,2]
   
    # Flight path angle
    gamma = theta-alpha

    # process
    v_x =   np.cos(beta) *v_mag * np.cos(gamma)
    v_y =   np.sin(beta) *v_mag * np.cos(gamma)
    v_z = -v_mag * np.sin(gamma) # z points down

    # pack
    conditions.frames.inertial.velocity_vector[:,0] = v_x[:,0]
    conditions.frames.inertial.velocity_vector[:,1] = v_y[:,0]
    conditions.frames.inertial.velocity_vector[:,2] = v_z[:,0]
 

def update_differentials_altitude(segment):
    """On each iteration creates the differentials and integration funcitons from knowns about the problem. Sets the time at each point. Must return in dimensional time, with t[0] = 0
    
    Assumptions:
    Constant throttle setting, with a constant true airspeed.

    Source:
    N/A

    Inputs:
    segment.climb_angle                         [radians]
    state.conditions.frames.inertial.velocity_vector [meter/second]
    segment.altitude_start                      [meters]
    segment.altitude_end                        [meters]

    Outputs:
    state.conditions.frames.inertial.time       [seconds]
    conditions.frames.inertial.position_vector  [meters]
    conditions.freestream.altitude              [meters]

    Properties Used:
    N/A
    """   

    # unpack
    t = segment.state.numerics.dimensionless.control_points
    I = segment.state.numerics.dimensionless.integrate
    
    # Unpack segment initials
    alt0       = segment.altitude_start 
    altf       = segment.altitude_end    
    conditions = segment.state.conditions  
    v          = segment.state.conditions.frames.inertial.velocity_vector
    
    # check for initial altitude
    if alt0 is None:
        if not segment.state.initials: raise AttributeError('initial altitude not set')
        alt0 = -1.0 *segment.state.initials.conditions.frames.inertial.position_vector[-1,2]    
    
    # get overall time step
    vz = -v[:,2,None] # Inertial velocity is z down
    dz = altf- alt0    
    dt = dz / np.dot(I[-1,:],vz)[-1] # maintain column array
    
    # Integrate vz to get altitudes
    alt = alt0 + np.dot(I*dt,vz)

    # rescale operators
    t = t * dt

    # pack
    t_initial = segment.state.conditions.frames.inertial.time[0,0]
    segment.state.conditions.frames.inertial.time[:,0] = t_initial + t[:,0]
    conditions.frames.inertial.position_vector[:,2]    = -alt[:,0] # z points down
    conditions.freestream.altitude[:,0]                =  alt[:,0] # positive altitude in this context    

    return

# ----------------------------------------------------------------------
#  Update Velocity Vector from Wind Angle
# ----------------------------------------------------------------------

def update_velocity_vector_from_wind_angle(segment):
    
    # unpack
    conditions = segment.state.conditions 
    v_mag      = segment.air_speed 
    beta       = segment.sideslip_angle
    alpha      = segment.state.unknowns.wind_angle[:,0][:,None]
    theta      = segment.state.unknowns.body_angle[:,0][:,None]
    
    # Flight path angle
    gamma = theta-alpha

    # process
    v_x =   np.cos(beta) *v_mag * np.cos(gamma)
    v_y =   np.sin(beta) *v_mag * np.cos(gamma)
    v_z = -v_mag * np.sin(gamma) # z points down

    # pack
    conditions.frames.inertial.velocity_vector[:,0] = v_x[:,0]
    conditions.frames.inertial.velocity_vector[:,1] = v_y[:,0]
    conditions.frames.inertial.velocity_vector[:,2] = v_z[:,0]

    return conditions
